import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel,conv2d,eig22, time_it

@time_it
def misomega(infolder, outname=None, chi_in=45, psi_in=0, window_size=1,write_flag=True,max_workers=None):
    input_filepaths = [
        os.path.join(infolder, "C11.bin"), 
        os.path.join(infolder, "C12_real.bin"),
        os.path.join(infolder, "C12_imag.bin"),
        os.path.join(infolder, "C22.bin")
    ]

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "Ps_miSOmega.tif"))
        output_filepaths.append(os.path.join(infolder, "Pd_miSOmega.tif"))
        output_filepaths.append(os.path.join(infolder, "Pv_miSOmega.tif"))
        

    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, 
                write_flag=write_flag,processing_func=process_chunk_misomega,block_size=(512, 512), 
                max_workers=max_workers,  num_outputs=3, chi_in=chi_in,psi_in=psi_in)

def process_chunk_misomega(chunks, window_size,input_filepaths,chi_in,psi_in):

    kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)
    c11_T1 = np.array(chunks[0])
    c12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
    c21_T1 = np.conj(c12_T1)
    c22_T1 = np.array(chunks[3])

    ncols,nrows = np.shape(c11_T1)

    if window_size>1:
        c11_T1 = conv2d(np.real(c11_T1),kernel)+1j*conv2d(np.imag(c11_T1),kernel)
        c12_T1 = conv2d(np.real(c12_T1),kernel)+1j*conv2d(np.imag(c12_T1),kernel)
        c21_T1 = conv2d(np.real(c21_T1),kernel)+1j*conv2d(np.imag(c21_T1),kernel)
        c22_T1 = conv2d(np.real(c22_T1),kernel)+1j*conv2d(np.imag(c22_T1),kernel)
        
    # Compute Stokes parameters
    s0 = np.abs(c11_T1 + c22_T1)
    s1 = np.abs(c11_T1 - c22_T1)
    s2 = np.abs(c12_T1 + c21_T1)
    s3 = np.where(chi_in >= 0, 1j * (c12_T1 - c21_T1), -1j * (c12_T1 - c21_T1))
    s3 = np.real(s3)

    ## Stokes child parameters
    SC = ((s0)-(s3))/2;
    OC = ((s0)+(s3))/2;
    
    CPR = np.divide(SC,OC)  ##SC/OC
    # CPR = SC/OC
    ##scattered fields    
    dop= np.sqrt(np.power(s1,2) + np.power(s2,2) + np.power(s3,2))/(s0)
    Psi = 0.5*((180/np.pi)*np.arctan2(s2,s1))
    DOCP = (-s3)/(dop*s0);
    Chi = 0.5*((180/np.pi)*np.arcsin(DOCP))
    ##---------------------------------
    
    ##---------------------------------
    # Calculating Omega from S-Omega decomposition        
    x1 = np.cos(2*chi_in*np.pi/180)*np.cos(2*psi_in*np.pi/180)*np.cos(2*Chi*np.pi/180)*np.cos(2*Psi*np.pi/180)
    x2 = np.cos(2*chi_in*np.pi/180)*np.sin(2*psi_in*np.pi/180)*np.cos(2*Chi*np.pi/180)*np.sin(2*Psi*np.pi/180)
    x3 = np.abs(np.sin(2*chi_in*np.pi/180)*np.sin(2*Chi*np.pi/180))
    Prec =  dop*(1 + x1 + x2 + x3)
    Prec1 = (1 - dop) + dop*(1 + x1 + x2 + x3)
    omega = (Prec/Prec1)
       
    
    # ## Improved S-Omega (i-SOmega powers
    ind_g1 = (CPR>1).astype(int)
    s_new_g1 = omega*(1 - omega)*OC
    db_new_g1 = omega*s0 - omega*(1 - omega)*OC   ##depolarized of OC x polarized of SC

    ind_l1 = (CPR<1).astype(int)
    s_new_l1 = omega*s0 - omega*(1 - omega)*SC
    db_new_l1 = omega*(1 - omega)*SC   ##depolarized of OC x polarized of SC

    ind_e1 = (CPR==1).astype(int)
    s_new_e1 = omega*OC
    db_new_e1 = omega*SC
                     
    
    surface_new = s_new_g1*ind_g1+s_new_l1*ind_l1+s_new_e1*ind_e1
    double_bounce_new = db_new_g1*ind_g1+db_new_l1*ind_l1+db_new_e1*ind_e1

    diffused_new = (1 - omega)*s0; ##diffused scattering
    
    surface_new[surface_new==0] = np.nan
    double_bounce_new[double_bounce_new==0] = np.nan
    diffused_new[diffused_new==0] = np.nan
    

    return surface_new, double_bounce_new, diffused_new


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it
from polsartools.cprvicpp import process_chunk_cprvicpp

@time_it
def cprvi(infolder, outname=None, chi_in=45, psi_in=0,window_size=1,write_flag=True,max_workers=None):
    input_filepaths = [
        os.path.join(infolder, "C11.bin"), 
        os.path.join(infolder, "C12_real.bin"),
        os.path.join(infolder, "C12_imag.bin"),
        os.path.join(infolder, "C22.bin")
    ]

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "cprvi.tif"))

    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, 
                write_flag=write_flag,processing_func=process_chunk_cprvi,block_size=(512, 512), 
                max_workers=max_workers,  num_outputs=1, chi_in=chi_in,psi_in=psi_in)

def process_chunk_cprvi(chunks, window_size,input_filepaths,chi_in,psi_in):

    chunk_arrays = [np.array(ch) for ch in chunks]  
    # CPP function
    vi_c_raw = process_chunk_cprvicpp( chunk_arrays, window_size, input_filepaths, chi_in, psi_in )

    return np.array(vi_c_raw, copy=True)  
    
 


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel,conv2d,eig22, time_it

@time_it
def mf3cc(infolder, outname=None, chi_in=45, psi_in=0, window_size=1,write_flag=True,max_workers=None):
    input_filepaths = [
        os.path.join(infolder, "C11.bin"), 
        os.path.join(infolder, "C12_real.bin"),
        os.path.join(infolder, "C12_imag.bin"),
        os.path.join(infolder, "C22.bin")
    ]

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "Ps_mf3cc.tif"))
        output_filepaths.append(os.path.join(infolder, "Pd_mf3cc.tif"))
        output_filepaths.append(os.path.join(infolder, "Pv_mf3cc.tif"))
        output_filepaths.append(os.path.join(infolder, "Theta_CP_mf3cc.tif"))
        

    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, 
                write_flag=write_flag,processing_func=process_chunk_mf3cc,block_size=(512, 512), 
                max_workers=max_workers,  num_outputs=4, chi_in=chi_in,psi_in=psi_in)

def process_chunk_mf3cc(chunks, window_size,input_filepaths,chi_in,psi_in):

    kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)
    c11_T1 = np.array(chunks[0])
    c12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
    c21_T1 = np.conj(c12_T1)
    c22_T1 = np.array(chunks[3])

    ncols,nrows = np.shape(c11_T1)

    if window_size>1:
        c11_T1 = conv2d(np.real(c11_T1),kernel)+1j*conv2d(np.imag(c11_T1),kernel)
        c12_T1 = conv2d(np.real(c12_T1),kernel)+1j*conv2d(np.imag(c12_T1),kernel)
        c21_T1 = conv2d(np.real(c21_T1),kernel)+1j*conv2d(np.imag(c21_T1),kernel)
        c22_T1 = conv2d(np.real(c22_T1),kernel)+1j*conv2d(np.imag(c22_T1),kernel)
    
    c2_det = (c11_T1*c22_T1-c12_T1*c21_T1)
    c2_trace = c11_T1+c22_T1
    # t2_span = t11s*t22s
    m1 = np.real(np.sqrt(1.0-(4.0*c2_det/np.power(c2_trace,2))))

    # Compute Stokes parameters
    s0 = c11_T1 + c22_T1
    s1 = c11_T1 - c22_T1
    s2 = np.real(c12_T1 + c21_T1)
    s3 = np.where(chi_in >= 0, 1j * (c12_T1 - c21_T1), -1j * (c12_T1 - c21_T1))
    s3 = np.real(s3)

    SC = ((s0)-(s3))/2;
    OC = ((s0)+(s3))/2;

    h = (OC-SC)
    # span = c11s + c22s

    val = ((m1*s0*h))/((SC*OC + (m1**2)*(s0**2)))
    thet = np.real(np.arctan(val))
    theta_CP = np.rad2deg(thet)

    Ps_CP= (((m1*(c2_trace)*(1.0+np.sin(2*thet))/2)))
    Pd_CP= (((m1*(c2_trace)*(1.0-np.sin(2*thet))/2)))
    Pv_CP= (c2_trace*(1.0-m1))
    

    return Ps_CP, Pd_CP, Pv_CP, theta_CP


import os

def get_filter_io_paths(infolder, outname, window_size, filter_type=None):
    # Determine the input filepaths based on available matrix types
    input_filepaths = []
    matrix_type = None  # To identify if it's C2, C3, T2, or T3

    if os.path.isfile(os.path.join(infolder, "T11.bin")) and os.path.isfile(os.path.join(infolder, "T33.bin")):
        matrix_type = "T3"
        input_filepaths = [
            os.path.join(infolder, "T11.bin"),
            os.path.join(infolder, 'T12_real.bin'), os.path.join(infolder, 'T12_imag.bin'),
            os.path.join(infolder, 'T13_real.bin'), os.path.join(infolder, 'T13_imag.bin'),
            os.path.join(infolder, "T22.bin"),
            os.path.join(infolder, 'T23_real.bin'), os.path.join(infolder, 'T23_imag.bin'),
            os.path.join(infolder, "T33.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder, "C11.bin")) and os.path.isfile(os.path.join(infolder, "C33.bin")):
        matrix_type = "C3"
        input_filepaths = [
            os.path.join(infolder, "C11.bin"),
            os.path.join(infolder, 'C12_real.bin'), os.path.join(infolder, 'C12_imag.bin'),
            os.path.join(infolder, 'C13_real.bin'), os.path.join(infolder, 'C13_imag.bin'),
            os.path.join(infolder, "C22.bin"),
            os.path.join(infolder, 'C23_real.bin'), os.path.join(infolder, 'C23_imag.bin'),
            os.path.join(infolder, "C33.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder, "C11.bin")) and os.path.isfile(os.path.join(infolder, "C22.bin")) \
            and not os.path.isfile(os.path.join(infolder, "C33.bin")):
        matrix_type = "C2"
        input_filepaths = [
            os.path.join(infolder, "C11.bin"),
            os.path.join(infolder, 'C12_real.bin'), os.path.join(infolder, 'C12_imag.bin'),
            os.path.join(infolder, "C22.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder, "T11.bin")) and os.path.isfile(os.path.join(infolder, "T22.bin")) \
            and not os.path.isfile(os.path.join(infolder, "T33.bin")):
        matrix_type = "T2"
        input_filepaths = [
            os.path.join(infolder, "T11.bin"),
            os.path.join(infolder, 'T12_real.bin'), os.path.join(infolder, 'T12_imag.bin'),
            os.path.join(infolder, "T22.bin"),
        ]
    else:
        raise ValueError("Invalid C2/C3 or T2/T3 folder!")

    # Determine the output folder and filepaths
    output_filepaths = []
    if outname is None:
        # outFolder = os.path.join(os.path.dirname(infolder), os.path.basename(infolder) + f"_{window_size}x{window_size}")
        outFolder = os.path.join(os.path.dirname(infolder)+ f"_rlee_{window_size}x{window_size}", os.path.basename(infolder) )
        os.makedirs(outFolder, exist_ok=True)
        
        # Only use the first letter of the matrix type (C for C3, T for T3)
        matrix_prefix = matrix_type[0]

        if matrix_type == "C3" or matrix_type == "T3":
            output_filepaths = [
                os.path.join(outFolder, f"{matrix_prefix}11.bin"),
                os.path.join(outFolder, f"{matrix_prefix}12_real.bin"), os.path.join(outFolder, f"{matrix_prefix}12_imag.bin"),
                os.path.join(outFolder, f"{matrix_prefix}13_real.bin"), os.path.join(outFolder, f"{matrix_prefix}13_imag.bin"),
                os.path.join(outFolder, f"{matrix_prefix}22.bin"),
                os.path.join(outFolder, f"{matrix_prefix}23_real.bin"), os.path.join(outFolder, f"{matrix_prefix}23_imag.bin"),
                os.path.join(outFolder, f"{matrix_prefix}33.bin"),
            ]
        elif matrix_type == "C2" or matrix_type == "T2":
            output_filepaths = [
                os.path.join(outFolder, f"{matrix_prefix}11.bin"),
                os.path.join(outFolder, f"{matrix_prefix}12_real.bin"), os.path.join(outFolder, f"{matrix_prefix}12_imag.bin"),
                os.path.join(outFolder, f"{matrix_prefix}22.bin"),
            ]
    
    return input_filepaths, output_filepaths



# polsartools/polsar/fp/__init__.py

from .rvifp import rvifp
from .mf3cf import mf3cf
from .mf4cf import mf4cf
from .dopfp import dopfp
from .prvifp import prvifp
from .grvi import grvi
from .neufp import neufp
from .nnedfp import nnedfp
from .halphafp import halphafp


import os
import numpy as np
from osgeo import gdal
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.utils.convert_matrices import T3_C3_mat

@time_it
def convert_T3_C3(infolder, outname=None, chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):

    if os.path.isfile(os.path.join(infolder,"T11.bin")):
        
        ds = gdal.Open(os.path.join(infolder,"T11.bin"))
        rows,cols = ds.RasterYSize, ds.RasterXSize 
        ds = None
        input_filepaths = [
        os.path.join(infolder,"T11.bin"),
        os.path.join(infolder,'T12_real.bin'), os.path.join(infolder,'T12_imag.bin'),  
        os.path.join(infolder,'T13_real.bin'), os.path.join(infolder,'T13_imag.bin'),
        os.path.join(infolder,"T22.bin"),
        os.path.join(infolder,'T23_real.bin'), os.path.join(infolder,'T23_imag.bin'),  
        os.path.join(infolder,"T33.bin"),
        ]
       
    # elif os.path.isfile(os.path.join(infolder,"C11.bin")):
    #     input_filepaths = [
    #     os.path.join(infolder,"C11.bin"),
    #     os.path.join(infolder,'C12_real.bin'), os.path.join(infolder,'C12_imag.bin'),  
    #     os.path.join(infolder,'C13_real.bin'), os.path.join(infolder,'C13_imag.bin'),
    #     os.path.join(infolder,"C22.bin"),
    #     os.path.join(infolder,'C23_real.bin'), os.path.join(infolder,'C23_imag.bin'),  
    #     os.path.join(infolder,"C33.bin"),
    #     ]
        os.makedirs(os.path.join(os.path.dirname(infolder), "C3"), exist_ok=True)
    else:
        raise Exception(f"Invalid T3 folder!!")

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C11.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C12_real.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C12_imag.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C13_real.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C13_imag.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C22.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C23_real.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C23_imag.bin"))
        output_filepaths.append(os.path.join(os.path.dirname(infolder), "C3", "C33.bin"))
    
    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
            processing_func=process_chunk_T3C3,
            block_size=(512, 512), max_workers=max_workers, 
            num_outputs=9)
    
    file = open(os.path.join(os.path.dirname(infolder), "C3",'config.txt') ,"w+")
    file.write('Nrow\n%d\n---------\nNcol\n%d\n---------\nPolarCase\nmonostatic\n---------\nPolarType\nfull'%(rows,cols))
    file.close()  

def process_chunk_T3C3(chunks, window_size, input_filepaths, *args):

    # additional_arg1 = args[0] if len(args) > 0 else None
    # additional_arg2 = args[1] if len(args) > 1 else None

    if 'T11' in input_filepaths[0] and 'T22' in input_filepaths[5] and 'T33' in input_filepaths[8]:
        
        t11_T1 = np.array(chunks[0])
        t12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
        t13_T1 = np.array(chunks[3])+1j*np.array(chunks[4])
        t21_T1 = np.conj(t12_T1)
        t22_T1 = np.array(chunks[5])
        t23_T1 = np.array(chunks[6])+1j*np.array(chunks[7])
        t31_T1 = np.conj(t13_T1)
        t32_T1 = np.conj(t23_T1)
        t33_T1 = np.array(chunks[8])

        T3 = np.array([[t11_T1, t12_T1, t13_T1], 
                     [t21_T1, t22_T1, t23_T1], 
                     [t31_T1, t32_T1, t33_T1]])
        T_T1 = T3_C3_mat(T3)
    else:
        raise Exception(f"Invalid T3 folder!!")

    # if 'C11' in input_filepaths[0] and 'C22' in input_filepaths[5] and 'C33' in input_filepaths[8]:
    #     C11 = np.array(chunks[0])
    #     C12 = np.array(chunks[1])+1j*np.array(chunks[2])
    #     C13 = np.array(chunks[3])+1j*np.array(chunks[4])
    #     C21 = np.conj(C12)
    #     C22 = np.array(chunks[5])
    #     C23 = np.array(chunks[6])+1j*np.array(chunks[7])
    #     C31 = np.conj(C13)
    #     C32 = np.conj(C23)
    #     C33 = np.array(chunks[8])
    #     T_T1 = np.array([[C11, C12, C13], 
    #                      [C21, C22, C23], 
    #                      [C31, C32, C33]])


    if window_size>1:
        kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)

        t11f = conv2d(T_T1[0,0,:,:],kernel)
        t12f = conv2d(np.real(T_T1[0,1,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,1,:,:]),kernel)
        t13f = conv2d(np.real(T_T1[0,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,2,:,:]),kernel)
        
        t21f = np.conj(t12f) 
        t22f = conv2d(T_T1[1,1,:,:],kernel)
        t23f = conv2d(np.real(T_T1[1,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[1,2,:,:]),kernel)

        t31f = np.conj(t13f) 
        t32f = np.conj(t23f) 
        t33f = conv2d(T_T1[2,2,:,:],kernel)

        T_T1 = np.array([[t11f, t12f, t13f], [t21f, t22f, t23f], [t31f, t32f, t33f]])


    return T_T1[0,0,:,:], T_T1[0,1,:,:].real, T_T1[0,1,:,:].imag, T_T1[0,2,:,:].real, T_T1[0,2,:,:].imag, T_T1[1,1,:,:], T_T1[1,2,:,:].real, T_T1[1,2,:,:].imag, T_T1[2,2,:,:]




import polsartools as pst

#%%
compact_c2 = r'./sample_data/compact_pol/C2_RHV'

cr = pst.polsar.cp.cprvi(compact_c2,chi_in=45,window_size=3,  write_flag=False)
dcp = pst.polsar.cp.dopcp(compact_c2,chi_in=45,window_size=3,  write_flag=False)
ccp = pst.polsar.cp.mf3cc(compact_c2,chi_in=45,window_size=3,  write_flag=False)
socp = pst.polsar.cp.misomega(compact_c2,chi_in=45,psi_in=0,window_size=3,  write_flag=False)
print('compact pol')
#%%
full_T3 = r'./sample_data/full_pol/T3'

cf3 = pst.polsar.fp.mf3cf(full_T3,window_size=3,write_flag=False)
cf4 = pst.polsar.fp.mf4cf(full_T3,window_size=3,write_flag=False)
dfp = pst.polsar.fp.dopfp(full_T3,window_size=3,write_flag=False)
gr = pst.polsar.fp.grvi(full_T3,window_size=3,write_flag=False)
rv = pst.polsar.fp.rvifp(full_T3,window_size=3,write_flag=False)
pr = pst.polsar.fp.prvifp(full_T3,window_size=3,write_flag=False)

print('full pol')

#%%

dxp_C2 = r'./sample_data/dual_pol/C2_VVVH'

dpr = pst.polsar.dxp.dprvi(dxp_C2,window_size=3,write_flag=False)
rvdp = pst.polsar.dxp.rvidp(dxp_C2,window_size=3,write_flag=False)
prvdp = pst.polsar.dxp.prvidp(dxp_C2,window_size=3,write_flag=False)
ddp = pst.polsar.dxp.dopdp(dxp_C2,window_size=3,write_flag=False)

print('dual cross-pol')




import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.utils.convert_matrices import T3_C3_mat

@time_it
def nnedfp(infolder, outname=None, chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):

    if os.path.isfile(os.path.join(infolder,"T11.bin")):
        input_filepaths = [
        os.path.join(infolder,"T11.bin"),
        os.path.join(infolder,'T12_real.bin'), os.path.join(infolder,'T12_imag.bin'),  
        os.path.join(infolder,'T13_real.bin'), os.path.join(infolder,'T13_imag.bin'),
        os.path.join(infolder,"T22.bin"),
        os.path.join(infolder,'T23_real.bin'), os.path.join(infolder,'T23_imag.bin'),  
        os.path.join(infolder,"T33.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder,"C11.bin")):
        input_filepaths = [
        os.path.join(infolder,"C11.bin"),
        os.path.join(infolder,'C12_real.bin'), os.path.join(infolder,'C12_imag.bin'),  
        os.path.join(infolder,'C13_real.bin'), os.path.join(infolder,'C13_imag.bin'),
        os.path.join(infolder,"C22.bin"),
        os.path.join(infolder,'C23_real.bin'), os.path.join(infolder,'C23_imag.bin'),  
        os.path.join(infolder,"C33.bin"),
        ]

    else:
        print(f"Invalid C3 or T3 folder!!")

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "NNED_odd.tif"))
        output_filepaths.append(os.path.join(infolder, "NNED_dbl.tif"))
        output_filepaths.append(os.path.join(infolder, "NNED_vol.tif"))
        
    
    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
            processing_func=process_chunk_nnedfp,
            block_size=(512, 512), max_workers=max_workers, 
            num_outputs=3)

def process_chunk_nnedfp(chunks, window_size, input_filepaths, *args):

    # additional_arg1 = args[0] if len(args) > 0 else None
    # additional_arg2 = args[1] if len(args) > 1 else None

    if 'T11' in input_filepaths[0] and 'T22' in input_filepaths[5] and 'T33' in input_filepaths[8]:
        t11_T1 = np.array(chunks[0])
        t12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
        t13_T1 = np.array(chunks[3])+1j*np.array(chunks[4])
        t21_T1 = np.conj(t12_T1)
        t22_T1 = np.array(chunks[5])
        t23_T1 = np.array(chunks[6])+1j*np.array(chunks[7])
        t31_T1 = np.conj(t13_T1)
        t32_T1 = np.conj(t23_T1)
        t33_T1 = np.array(chunks[8])

        T3 = np.array([[t11_T1, t12_T1, t13_T1], 
                     [t21_T1, t22_T1, t23_T1], 
                     [t31_T1, t32_T1, t33_T1]])
        T_T1 = T3_C3_mat(T3)


    if 'C11' in input_filepaths[0] and 'C22' in input_filepaths[5] and 'C33' in input_filepaths[8]:
        C11 = np.array(chunks[0])
        C12 = np.array(chunks[1])+1j*np.array(chunks[2])
        C13 = np.array(chunks[3])+1j*np.array(chunks[4])
        C21 = np.conj(C12)
        C22 = np.array(chunks[5])
        C23 = np.array(chunks[6])+1j*np.array(chunks[7])
        C31 = np.conj(C13)
        C32 = np.conj(C23)
        C33 = np.array(chunks[8])
        T_T1 = np.array([[C11, C12, C13], 
                         [C21, C22, C23], 
                         [C31, C32, C33]])


    if window_size>1:
        kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)

        t11f = conv2d(T_T1[0,0,:,:],kernel)
        t12f = conv2d(np.real(T_T1[0,1,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,1,:,:]),kernel)
        t13f = conv2d(np.real(T_T1[0,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,2,:,:]),kernel)
        
        t21f = np.conj(t12f) 
        t22f = conv2d(T_T1[1,1,:,:],kernel)
        t23f = conv2d(np.real(T_T1[1,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[1,2,:,:]),kernel)

        t31f = np.conj(t13f) 
        t32f = np.conj(t23f) 
        t33f = conv2d(T_T1[2,2,:,:],kernel)

        T_T1 = np.array([[t11f, t12f, t13f], [t21f, t22f, t23f], [t31f, t32f, t33f]])

    

    _,_,rows,cols = np.shape(T_T1)
    T_T1 = T_T1.reshape(9, rows, cols)
    
    # Indices for vectorized access
    i, j = np.indices((rows, cols))

    SpanMax = -float('inf')  
    SpanMin = float('inf')  

    Span_data = np.real(T_T1[0,:,:]) + np.real(T_T1[4,:,:]) + np.real(T_T1[8,:,:])

    # Find the maximum and minimum values in one step
    SpanMax = np.nanmax(Span_data)
    SpanMin = np.nanmin(Span_data)

    # Ensure SpanMin does not go below eps
    SpanMin = np.nanmax([SpanMin, 1e-6])

    # Set all values in Span_data below eps to eps
    Span_data[Span_data < 1e-6] = 1e-6

    epsilon = np.real(T_T1[0,:,:])
    rho_re = np.real(T_T1[2,:,:])
    rho_im = np.imag(T_T1[2,:,:])
    nhu = np.real(T_T1[4,:,:])
    gamma = np.real(T_T1[8,:,:])
    
    # Pre-calculate epsilon_veg, rho_re_veg, rho_im_veg, nhu_veg, gamma_veg
    fv = 3. * nhu / 2.
    epsilon_veg = fv
    rho_re_veg = fv / 3.
    rho_im_veg = 0.
    nhu_veg = 2. * fv / 3.
    gamma_veg = fv
    
    # Calculate z, a, b
    z = epsilon * gamma_veg + epsilon_veg * gamma - 2. * rho_re * rho_re_veg
    a = epsilon_veg * gamma_veg - rho_re_veg * rho_re_veg
    b = epsilon * gamma - rho_re * rho_re - rho_im * rho_im

    # Calculate x1
    x1 = nhu / nhu_veg

    # Handle the case where a == 0 using a mask
    mask_a_zero = a == 0.
    x2 = np.where(mask_a_zero, b / z, (z - np.sqrt(z * z - 4. * a * b)) / (2. * a))
    xmax = np.minimum(x1, x2)
    
    # make sure the array is writable
    epsilon.setflags(write=True)  
    rho_re.setflags(write=True)
    rho_im.setflags(write=True)
    nhu.setflags(write=True)
    gamma.setflags(write=True)

    # Update epsilon, rho_re, rho_im, nhu, gamma
    epsilon -= xmax * epsilon_veg
    rho_re -= xmax * rho_re_veg
    rho_im -= xmax * rho_im_veg
    nhu -= xmax * nhu_veg
    gamma -= xmax * gamma_veg
    

    # Calculate delta, lambda1, lambda2
    delta = (epsilon - gamma) ** 2 + 4. * (rho_re * rho_re + rho_im * rho_im)
    lambda1 = 0.5 * (epsilon + gamma + np.sqrt(delta))
    lambda2 = 0.5 * (epsilon + gamma - np.sqrt(delta))

    # Calculate OMEGA1, OMEGA2
    OMEGA1 = lambda1 * (gamma - epsilon + np.sqrt(delta)) ** 2
    OMEGA1 /= ( (gamma - epsilon + np.sqrt(delta)) ** 2 + 4. * (rho_re * rho_re + rho_im * rho_im))

    OMEGA2 = lambda2 * (gamma - epsilon - np.sqrt(delta)) ** 2
    OMEGA2 /= ((gamma - epsilon - np.sqrt(delta)) ** 2 + 4. * (rho_re * rho_re + rho_im * rho_im))

    # Calculate hh1_re, hh1_im, hh2_re, hh2_im
    hh1_re = 2. * rho_re / (gamma - epsilon + np.sqrt(delta))
    hh1_im = 2. * rho_im / (gamma - epsilon + np.sqrt(delta))

    hh2_re = 2. * rho_re / (gamma - epsilon - np.sqrt(delta))
    hh2_im = 2. * rho_im / (gamma - epsilon - np.sqrt(delta))

    # Define A0A0 and B0pB
    A0A0 = (hh1_re + 1.) ** 2 + hh1_im ** 2
    B0pB = (hh1_re - 1.) ** 2 + hh1_im ** 2
    
    # Apply conditions for ALPre, ALPim, OMEGAodd, BETre, BETim, OMEGAdbl
    mask_A0A0_greater = A0A0 > B0pB
    ALPre = np.where(mask_A0A0_greater, hh1_re, hh2_re)
    ALPim = np.where(mask_A0A0_greater, hh1_im, hh2_im)
    OMEGAodd = np.where(mask_A0A0_greater, OMEGA1, OMEGA2)

    BETre = np.where(mask_A0A0_greater, hh2_re, hh1_re)
    BETim = np.where(mask_A0A0_greater, hh2_im, hh1_im)
    OMEGAdbl = np.where(mask_A0A0_greater, OMEGA2, OMEGA1)

    # Calculate NNED_odd, NNED_dbl, and NNED_vol
    NNED_odd = OMEGAodd * (1 + ALPre**2 + ALPim**2)
    NNED_dbl = OMEGAdbl * (1 + BETre**2 + BETim**2)
    NNED_vol = xmax * (epsilon_veg + nhu_veg + gamma_veg)

    # Apply limits to M_odd, M_dbl, M_vol
    NNED_odd = np.clip(NNED_odd, 0., SpanMax)
    NNED_dbl = np.clip(NNED_dbl, 0., SpanMax)
    NNED_vol = np.clip(NNED_vol, 0., SpanMax)
    
    return NNED_odd,NNED_dbl,NNED_vol


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d, eig22

@time_it
def halphadp(infolder, outname=None,  chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):
    input_filepaths = [
        os.path.join(infolder, "C11.bin"), 
        os.path.join(infolder, "C12_real.bin"),
        os.path.join(infolder, "C12_imag.bin"),
        os.path.join(infolder, "C22.bin")
    ]
    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "Hdp.tif"))
        output_filepaths.append(os.path.join(infolder, "alphadp.tif"))

    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,processing_func=process_chunk_halphadp,block_size=(512, 512), max_workers=max_workers,  num_outputs=2)

def process_chunk_halphadp(chunks, window_size,*args):
    kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)
    c11_T1 = np.array(chunks[0])
    c12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
    c21_T1 = np.conj(c12_T1)
    c22_T1 = np.array(chunks[3])

    # C2_stack = np.zeros((np.shape(c11_T1)[0],np.shape(c11_T1)[1],4))
    C2_stack = np.dstack((c11_T1,c12_T1,np.conj(c12_T1),c22_T1)).astype(np.complex64)

    if window_size>1:
        C2_stack[:,:,0] = conv2d(np.real(c11_T1),kernel)+1j*conv2d(np.imag(c11_T1),kernel)
        C2_stack[:,:,1] = conv2d(np.real(c12_T1),kernel)+1j*conv2d(np.imag(c12_T1),kernel)
        C2_stack[:,:,2] = conv2d(np.real(c21_T1),kernel)+1j*conv2d(np.imag(c21_T1),kernel)
        C2_stack[:,:,3] = conv2d(np.real(c22_T1),kernel)+1j*conv2d(np.imag(c22_T1),kernel)

    data = C2_stack.reshape( C2_stack.shape[0]*C2_stack.shape[1], C2_stack.shape[2] ).reshape((-1,2,2))

    evals, evecs = np.linalg.eig(data)
    
    
    evals[:,0][evals[:,0] <0] = 0
    evals[:,1][evals[:,1] >1] = 1
    
    eval_norm1 = (evals[:,1])/(evals[:,0] + evals[:,1])
    eval_norm1[eval_norm1<0]=0
    eval_norm1[eval_norm1>1]=1
    
    
    eval_norm2 = (evals[:,0])/(evals[:,0] + evals[:,1])
    eval_norm2[eval_norm2<0]=0
    eval_norm2[eval_norm2>1]=1
    
    # # %Alpha 1
    eig_vec_r1 = np.real(evecs[:,0,1])
    eig_vec_c1 = np.imag(evecs[:,0,1])
    alpha1 = np.arccos(np.sqrt(eig_vec_r1*eig_vec_r1 + eig_vec_c1*eig_vec_c1))*180/np.pi
    
    # # %Alpha 2
    eig_vec_r2 = np.real(evecs[:,0,0])
    eig_vec_c2 = np.imag(evecs[:,0,0])
    alpha2 = np.arccos(np.sqrt(eig_vec_r2*eig_vec_r2 + eig_vec_c2*eig_vec_c2))*180/np.pi
    
    # # %Cloude Alpha
    alpha_ = (eval_norm1*alpha1 + eval_norm2*alpha2)
    alpha_ = alpha_.reshape(C2_stack.shape[0],C2_stack.shape[1])
    # # %Entropy
    H = - eval_norm1*np.log10(eval_norm1)/np.log10(2) - eval_norm2*np.log10(eval_norm2)/np.log10(2)
    H = H.reshape(C2_stack.shape[0],C2_stack.shape[1])

    # alpha1 = alpha1.reshape(C2_stack.shape[0],C2_stack.shape[1])
    # alpha2 = alpha2.reshape(C2_stack.shape[0],C2_stack.shape[1])

    print(np.namean(H))

    return H,alpha_


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.utils.convert_matrices import C3_T3_mat

@time_it
def neufp(infolder, outname=None, chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):

    if os.path.isfile(os.path.join(infolder,"T11.bin")):
        input_filepaths = [
        os.path.join(infolder,"T11.bin"),
        os.path.join(infolder,'T12_real.bin'), os.path.join(infolder,'T12_imag.bin'),  
        os.path.join(infolder,'T13_real.bin'), os.path.join(infolder,'T13_imag.bin'),
        os.path.join(infolder,"T22.bin"),
        os.path.join(infolder,'T23_real.bin'), os.path.join(infolder,'T23_imag.bin'),  
        os.path.join(infolder,"T33.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder,"C11.bin")):
        input_filepaths = [
        os.path.join(infolder,"C11.bin"),
        os.path.join(infolder,'C12_real.bin'), os.path.join(infolder,'C12_imag.bin'),  
        os.path.join(infolder,'C13_real.bin'), os.path.join(infolder,'C13_imag.bin'),
        os.path.join(infolder,"C22.bin"),
        os.path.join(infolder,'C23_real.bin'), os.path.join(infolder,'C23_imag.bin'),  
        os.path.join(infolder,"C33.bin"),
        ]

    else:
        print(f"Invalid C3 or T3 folder!!")

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "Neu_psi.tif"))
        output_filepaths.append(os.path.join(infolder, "Neu_delta_mod.tif"))
        output_filepaths.append(os.path.join(infolder, "Neu_delta_pha.tif"))
        output_filepaths.append(os.path.join(infolder, "Neu_tau.tif"))
    
    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
            processing_func=process_chunk_neufp,
            block_size=(512, 512), max_workers=max_workers, 
            num_outputs=4)

def process_chunk_neufp(chunks, window_size, input_filepaths, *args):

    # additional_arg1 = args[0] if len(args) > 0 else None
    # additional_arg2 = args[1] if len(args) > 1 else None

    if 'T11' in input_filepaths[0] and 'T22' in input_filepaths[5] and 'T33' in input_filepaths[8]:
        t11_T1 = np.array(chunks[0])
        t12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
        t13_T1 = np.array(chunks[3])+1j*np.array(chunks[4])
        t21_T1 = np.conj(t12_T1)
        t22_T1 = np.array(chunks[5])
        t23_T1 = np.array(chunks[6])+1j*np.array(chunks[7])
        t31_T1 = np.conj(t13_T1)
        t32_T1 = np.conj(t23_T1)
        t33_T1 = np.array(chunks[8])

        T_T1 = np.array([[t11_T1, t12_T1, t13_T1], 
                     [t21_T1, t22_T1, t23_T1], 
                     [t31_T1, t32_T1, t33_T1]])


    if 'C11' in input_filepaths[0] and 'C22' in input_filepaths[5] and 'C33' in input_filepaths[8]:
        C11 = np.array(chunks[0])
        C12 = np.array(chunks[1])+1j*np.array(chunks[2])
        C13 = np.array(chunks[3])+1j*np.array(chunks[4])
        C21 = np.conj(C12)
        C22 = np.array(chunks[5])
        C23 = np.array(chunks[6])+1j*np.array(chunks[7])
        C31 = np.conj(C13)
        C32 = np.conj(C23)
        C33 = np.array(chunks[8])
        C3 = np.array([[C11, C12, C13], 
                         [C21, C22, C23], 
                         [C31, C32, C33]])

        T_T1 = C3_T3_mat(C3)


    if window_size>1:
        kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)

        t11f = conv2d(T_T1[0,0,:,:],kernel)
        t12f = conv2d(np.real(T_T1[0,1,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,1,:,:]),kernel)
        t13f = conv2d(np.real(T_T1[0,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,2,:,:]),kernel)
        
        t21f = np.conj(t12f) 
        t22f = conv2d(T_T1[1,1,:,:],kernel)
        t23f = conv2d(np.real(T_T1[1,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[1,2,:,:]),kernel)

        t31f = np.conj(t13f) 
        t32f = np.conj(t23f) 
        t33f = conv2d(T_T1[2,2,:,:],kernel)

        T_T1 = np.array([[t11f, t12f, t13f], [t21f, t22f, t23f], [t31f, t32f, t33f]])

    

    _,_,rows,cols = np.shape(T_T1)
    T_T1 = T_T1.reshape(9, rows, cols)
    
    # Indices for vectorized access
    i, j = np.indices((rows, cols))

    # Extract components of T3_stack once
    T3_stack_real = np.real(T_T1[:,i, j ])
    T3_stack_imag = np.imag(T_T1[:, i, j])

    T_00_0 = T3_stack_real[0, :, :]
    T_01_0 = T3_stack_real[1, :, :]
    T_01_1 = T3_stack_imag[1, :, :]
    T_02_0 = T3_stack_real[2, :, :]
    T_02_1 = T3_stack_imag[2, :, :]
    T_11_0 = T3_stack_real[4, :, :]
    T_12_0 = T3_stack_real[5, :, :]
    T_22_0 = T3_stack_real[8, :, :]

    # Compute the Phi value using the given formula
    Phi = 0.25 * (np.pi + np.arctan2(-2. * T_12_0, T_22_0 - T_11_0))

    # Adjust Phi for the correct orientation
    Phi[Phi <= np.pi / 4.] = Phi[Phi <= np.pi / 4.]
    Phi[Phi > np.pi / 4.] -= np.pi / 2.

    # Convert Phi to degrees
    Neumann_psi = Phi * 180. / np.pi

    # Pre-compute trigonometric terms
    cos_2Phi = np.cos(2 * Phi)
    sin_2Phi = np.sin(2 * Phi)
    sin_4Phi = np.sin(4 * Phi)

    # Coherency Matrix de-orientation
    T110 = T_00_0
    T12re0 = T_01_0 * cos_2Phi - T_02_0 * sin_2Phi
    T12im0 = T_01_1 * cos_2Phi - T_02_1 * sin_2Phi
    T220 = T_11_0 * cos_2Phi**2 - T_12_0 * sin_4Phi + T_22_0 * sin_2Phi**2
    T330 = T_11_0 * sin_2Phi**2 - T_12_0 * sin_4Phi + T_22_0 * cos_2Phi**2

    # Compute Neumann_delta_mod, Neumann_delta_pha, and Neumann_tau
    Neumann_delta_mod = np.sqrt((T220 + T330) / (T110 + np.finfo(float).eps))
    Neumann_delta_pha = np.arctan2(T12im0, T12re0) * 180. / np.pi
    Neumann_tau = 1. - ((np.sqrt(T12re0 ** 2 + T12im0 ** 2) / T110) / (Neumann_delta_mod + np.finfo(float).eps))

    return Neumann_psi, Neumann_delta_mod, Neumann_delta_pha,Neumann_tau


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.preprocessing.pre_utils import get_filter_io_paths

from polsartools.preprocessing.rflee_filter import process_chunk_refined_lee
from polsartools.rflee import process_chunk_rfleecpp
@time_it
def boxcar(infolder, outname=None, chi_in=0, psi_in=0, window_size=3, write_flag=True, max_workers=None):
    
    input_filepaths, output_filepaths = get_filter_io_paths(infolder, outname, window_size, filter_type="boxcar")

    # Process chunks in parallel
    num_outputs = len(output_filepaths)
    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
                            processing_func=process_chunk_boxcar, block_size=(512, 512), max_workers=max_workers, 
                            num_outputs=num_outputs)

def process_chunk_boxcar(chunks, window_size, input_filepaths, *args):
    filtered_chunks = []
    for i in range(len(chunks)):
        img = np.array(chunks[i])
        kernel = np.ones((window_size, window_size), np.float32) / (window_size * window_size)
        filtered_chunks.append(conv2d(img, kernel))
    return filtered_chunks


@time_it
def rlee(infolder, outname=None, chi_in=0, psi_in=0, window_size=3, write_flag=True, max_workers=None):
    # File reading and output setup similar to the boxcar filter
    input_filepaths, output_filepaths = get_filter_io_paths(infolder, outname, window_size, filter_type="refined_lee")
    num_outputs = len(output_filepaths)
    
    ### Python implementation
    # process_chunks_parallel(input_filepaths, output_filepaths, window_size=window_size, write_flag=write_flag,
    #                         processing_func=process_chunk_refined_lee, block_size=(512, 512), max_workers=max_workers,
    #                         num_outputs=num_outputs)
    
    #### Uncomment below to use C++ implementation 

    process_chunks_parallel(input_filepaths, output_filepaths, window_size=window_size, write_flag=write_flag,
                        processing_func=process_chunk_rfl, block_size=(512, 512), max_workers=max_workers,
                        num_outputs=num_outputs)

def process_chunk_rfl(chunks, window_size,input_filepaths, *args):

    # print('before pad',np.shape(chunks[0]))
    
    for i in range(len(chunks)):
        pad_top_left = window_size // 2 
        pad_bottom_right = window_size // 2 +1
        chunks[i] = np.pad(chunks[i], 
                            ((pad_top_left, pad_bottom_right), 
                            (pad_top_left, pad_bottom_right)), 
                            mode='constant', constant_values=0)
    
    
    # print("after pad",np.shape(chunks[0]))
    chunk_arrays = [np.array(ch) for ch in chunks]  
    vi_c_raw = process_chunk_rfleecpp(chunk_arrays, window_size)
    
    proc_chunks=[]
    for chunk in vi_c_raw:
        proc_chunks.append(np.array(chunk))
        # print('mean %0.3f'%np.nanmean(np.array(chunk)),'std %0.3f'%np.nanstd(np.array(chunk)))
        
        
    del vi_c_raw
    
    # print("proc_chunks pad",np.shape(proc_chunks[0]))
    for i in range(len(proc_chunks)):
        # Calculate the padding size
        pad_top_left = window_size // 2 
        pad_bottom_right = window_size // 2 +1
        proc_chunks[i] = proc_chunks[i][pad_top_left:-pad_bottom_right, pad_top_left:-pad_bottom_right]
    
    
    def shift_array(arr,shift):
        # Get the number of rows and columns in the array
        rows, cols = arr.shape

        # Step 1: Move the rightmost 3 columns to the left
        right_columns = arr[:, -shift:]  # Last 3 columns
        remaining_columns = arr[:, :-shift]  # All but last 3 columns

        # Step 2: Move the bottom 3 rows to the top
        bottom_rows = arr[-shift:, :]  # Last 3 rows
        remaining_rows = arr[:-shift, :]  # All but last 3 rows

        # Combine the shifted rows and columns
        shifted_array = np.vstack((bottom_rows, remaining_rows))  # Stack bottom rows to the top
        shifted_array = np.hstack((right_columns, shifted_array[:, :-shift]))  # Stack right columns to the left

        return shifted_array
    
    
    for i in range(len(proc_chunks)):
        # Remove 'window_size' rows from the top and 'window_size' columns from the left
        proc_chunks[i] = shift_array(proc_chunks[i],window_size//2)
    
    
    # print("proc_chunks unpad",np.shape(proc_chunks[0]))

    
    num_chunks = len(proc_chunks) // 2
    out_chunks = []

    for i in range(num_chunks):
        real_part = proc_chunks[2 * i]       # Get the real part from the even indices
        imag_part = proc_chunks[2 * i + 1]   # Get the imaginary part from the odd indices
        complex_array = real_part + 1j * imag_part  # Create a complex number
        out_chunks.append(complex_array)
        # print(np.nanmean(real_part),' ' ,np.nanmean(imag_part))
    filtered_chunks = []

    if len(chunks)==9:
        # print("out_chunks shape:", np.shape(out_chunks))
        filtered_chunks.append(np.real(out_chunks[0]))
        filtered_chunks.append(np.real(out_chunks[1]))
        filtered_chunks.append(np.imag(out_chunks[1]))
        filtered_chunks.append(np.real(out_chunks[2]))
        filtered_chunks.append(np.imag(out_chunks[2]))
        filtered_chunks.append(np.real(out_chunks[4]))
        filtered_chunks.append(np.real(out_chunks[5]))
        filtered_chunks.append(np.imag(out_chunks[5]))
        filtered_chunks.append(np.real(out_chunks[8]))
    if len(chunks)==4:
        filtered_chunks.append(np.real(out_chunks[0]))
        filtered_chunks.append(np.real(out_chunks[1]))
        filtered_chunks.append(np.imag(out_chunks[1]))
        filtered_chunks.append(np.real(out_chunks[3]))
   
    return filtered_chunks
    


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 26 16:56:57 2025

@author: nbhogapurapu
"""



import polsartools as pst

#%%
def main():
    compact_c2 = r'tests/sample_data/compact_pol/C2_RHV'
    compact_c2 =r"C://Users/nbhogapurapu/Desktop/temp\ALOS2_read\data/C2"

    cr = pst.polsar.cp.cprvi(compact_c2,chi_in=45,window_size=7,  write_flag=True)
    # cr = pst.polsar.cp.cprvi(compact_c2,chi_in=45,window_size=7,  write_flag=True)
    # cr = pst.polsar.cp.cprvi(compact_c2,chi_in=45,window_size=7,  write_flag=True)
    # cr = pst.polsar.cp.cprvi(compact_c2,chi_in=45,window_size=7,  write_flag=True)
    # cr = pst.polsar.cp.cprvi(compact_c2,chi_in=45,window_size=7,  write_flag=True)
    
# dcp = pst.polsar.cp.dopcp(compact_c2,chi_in=45,window_size=3,  write_flag=False)
# ccp = pst.polsar.cp.mf3cc(compact_c2,chi_in=45,window_size=3,  write_flag=False)
# socp = pst.polsar.cp.misomega(compact_c2,chi_in=45,psi_in=0,window_size=3,  write_flag=False)
# print('compact pol')
# #%%
# full_T3 = r'./sample_data/full_pol/T3'

# cf3 = pst.polsar.fp.mf3cf(full_T3,window_size=3,write_flag=False)
# cf4 = pst.polsar.fp.mf4cf(full_T3,window_size=3,write_flag=False)
# dfp = pst.polsar.fp.dopfp(full_T3,window_size=3,write_flag=False)
# gr = pst.polsar.fp.grvi(full_T3,window_size=3,write_flag=False)
# rv = pst.polsar.fp.rvifp(full_T3,window_size=3,write_flag=False)
# pr = pst.polsar.fp.prvifp(full_T3,window_size=3,write_flag=False)

# print('full pol')

# #%%

# dxp_C2 = r'./sample_data/dual_pol/C2_VVVH'

# dpr = pst.polsar.dxp.dprvi(dxp_C2,window_size=3,write_flag=False)
# rvdp = pst.polsar.dxp.rvidp(dxp_C2,window_size=3,write_flag=False)
# prvdp = pst.polsar.dxp.prvidp(dxp_C2,window_size=3,write_flag=False)
# ddp = pst.polsar.dxp.dopdp(dxp_C2,window_size=3,write_flag=False)

# print('dual cross-pol')

if __name__ == '__main__':
    main()


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.utils.convert_matrices import C3_T3_mat
from osgeo import gdal
import matplotlib.pyplot as plt

def read_bin(file):
    ds = gdal.Open(file)
    band = ds.GetRasterBand(1)
    arr = band.ReadAsArray()
    return arr

def norm_data(data):
    
    data = 10*np.log10(data)
    data[data==-np.inf] = np.nan
    data[data==np.inf] = np.nan
    data = (data - np.nanmin(data)) / (np.nanmax(data) - np.nanmin(data))

    p5 = np.nanpercentile(data, 5)
    p95 = np.nanpercentile(data, 95)
    data = np.clip(data, p5, p95)
    data = (data - p5) / (p95 - p5)

    return data   

@time_it
def pauliRGB(infolder, outname=None, chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):
    if os.path.isfile(os.path.join(infolder,"T11.bin")) and os.path.isfile(os.path.join(infolder,"T22.bin")) and os.path.isfile(os.path.join(infolder,"T33.bin")):
        blue_ = norm_data(read_bin(os.path.join(infolder,"T11.bin")))
        red_ = norm_data(read_bin(os.path.join(infolder,"T22.bin")))
        green_ = norm_data(read_bin(os.path.join(infolder,"T33.bin")))

        rgb_uint8 = (np.dstack((red_,green_,blue_)) * 255) .astype(np.uint8)

        plt.imsave(os.path.join(infolder,"PauliRGB.png"),rgb_uint8)
        print(f"Pauli RGB image saved as {os.path.join(infolder,"PauliRGB.png")}PauliRGB.png")
        fig,ax = plt.subplots()
        plt.imshow(rgb_uint8,vmin=0,vmax=255)
        ax.axis('off')
        plt.savefig(os.path.join(infolder,"PauliRGB_thumb.png"), format='png', bbox_inches='tight', pad_inches=0)
    elif os.path.isfile(os.path.join(infolder,"C11.bin")) and os.path.isfile(os.path.join(infolder,"C22.bin")) and os.path.isfile(os.path.join(infolder,"C33.bin")):
        blue_ = norm_data(0.5*(read_bin(os.path.join(infolder,'C11.bin'))+read_bin(os.path.join(infolder,'C33.bin'))+2*read_bin(os.path.join(infolder,'C13_real.bin'))))
        red_ = norm_data(0.5*(read_bin(os.path.join(infolder,'C11.bin'))+read_bin(os.path.join(infolder,'C33.bin'))-2*read_bin(os.path.join(infolder,'C13_real.bin'))))
        green_ = norm_data(read_bin(os.path.join(infolder,"C22.bin")))

        rgb_uint8 = (np.dstack((red_,green_,blue_)) * 255) .astype(np.uint8)

        plt.imsave(os.path.join(infolder,"PauliRGB.png"),rgb_uint8)
        print(f"Pauli RGB image saved as {os.path.join(infolder,"PauliRGB.png")}PauliRGB.png")
        fig,ax = plt.subplots()
        plt.imshow(rgb_uint8,vmin=0,vmax=255)
        ax.axis('off')
        plt.savefig(os.path.join(infolder,"PauliRGB_thumb.png"), format='png', bbox_inches='tight', pad_inches=0)
    else:
        raise ValueError("Invalid C3/T3 folder!!")
        
        
        
        
        
        
        
# @time_it
# def pauliRGB(infolder, outname=None, chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):

#     if os.path.isfile(os.path.join(infolder,"T11.bin")):
#         input_filepaths = [
#         os.path.join(infolder,"T11.bin"),
#         os.path.join(infolder,'T12_real.bin'), os.path.join(infolder,'T12_imag.bin'),  
#         os.path.join(infolder,'T13_real.bin'), os.path.join(infolder,'T13_imag.bin'),
#         os.path.join(infolder,"T22.bin"),
#         os.path.join(infolder,'T23_real.bin'), os.path.join(infolder,'T23_imag.bin'),  
#         os.path.join(infolder,"T33.bin"),
#         ]
#     elif os.path.isfile(os.path.join(infolder,"C11.bin")):
#         input_filepaths = [
#         os.path.join(infolder,"C11.bin"),
#         os.path.join(infolder,'C12_real.bin'), os.path.join(infolder,'C12_imag.bin'),  
#         os.path.join(infolder,'C13_real.bin'), os.path.join(infolder,'C13_imag.bin'),
#         os.path.join(infolder,"C22.bin"),
#         os.path.join(infolder,'C23_real.bin'), os.path.join(infolder,'C23_imag.bin'),  
#         os.path.join(infolder,"C33.bin"),
#         ]

#     else:
#         print(f"Invalid C3 or T3 folder!!")

#     output_filepaths = []
#     if outname is None:
#         output_filepaths.append(os.path.join(infolder, "red.bin"))
#         output_filepaths.append(os.path.join(infolder, "green.bin"))
#         output_filepaths.append(os.path.join(infolder, "blue.bin"))
    
#     process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
#             processing_func=process_chunk_pauliRGB,
#             block_size=(512, 512), max_workers=max_workers, 
#             num_outputs=3)

    
#     rgb_uint8 = np.dstack((read_bin(output_filepaths[0]),
#                            read_bin(output_filepaths[1]),
#                            read_bin(output_filepaths[2]))).astype(np.uint8)
    
#     plt.imsave(os.path.join(infolder,'PauliRGB.png'),rgb_uint8) 
#     # # plt.close(0)
#     fig,ax = plt.subplots()
#     plt.imshow(rgb_uint8,vmin=0,vmax=255)
#     ax.axis('off')
#     plt.savefig(os.path.join(infolder,'PauliRGB_thumb.png'), format='png', bbox_inches='tight', pad_inches=0)
#     plt.close()
    
#     os.remove(output_filepaths[0])
#     os.remove(output_filepaths[1])
#     os.remove(output_filepaths[2])
# def process_chunk_pauliRGB(chunks, window_size, input_filepaths, *args):

#     # additional_arg1 = args[0] if len(args) > 0 else None
#     # additional_arg2 = args[1] if len(args) > 1 else None

#     if 'T11' in input_filepaths[0] and 'T22' in input_filepaths[5] and 'T33' in input_filepaths[8]:
#         t11_T1 = np.array(chunks[0])
#         t12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
#         t13_T1 = np.array(chunks[3])+1j*np.array(chunks[4])
#         t21_T1 = np.conj(t12_T1)
#         t22_T1 = np.array(chunks[5])
#         t23_T1 = np.array(chunks[6])+1j*np.array(chunks[7])
#         t31_T1 = np.conj(t13_T1)
#         t32_T1 = np.conj(t23_T1)
#         t33_T1 = np.array(chunks[8])

#         T_T1 = np.array([[t11_T1, t12_T1, t13_T1], 
#                      [t21_T1, t22_T1, t23_T1], 
#                      [t31_T1, t32_T1, t33_T1]])
#         print(T_T1.shape)

#     if 'C11' in input_filepaths[0] and 'C22' in input_filepaths[5] and 'C33' in input_filepaths[8]:
#         C11 = np.array(chunks[0])
#         C12 = np.array(chunks[1])+1j*np.array(chunks[2])
#         C13 = np.array(chunks[3])+1j*np.array(chunks[4])
#         C21 = np.conj(C12)
#         C22 = np.array(chunks[5])
#         C23 = np.array(chunks[6])+1j*np.array(chunks[7])
#         C31 = np.conj(C13)
#         C32 = np.conj(C23)
#         C33 = np.array(chunks[8])
#         C3 = np.array([[C11, C12, C13], 
#                          [C21, C22, C23], 
#                          [C31, C32, C33]])

#         T_T1 = C3_T3_mat(C3)


#     if window_size>1:
#         kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)

#         t11f = conv2d(T_T1[0,0,:,:],kernel)
#         t12f = conv2d(np.real(T_T1[0,1,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,1,:,:]),kernel)
#         t13f = conv2d(np.real(T_T1[0,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,2,:,:]),kernel)
        
#         t21f = np.conj(t12f) 
#         t22f = conv2d(T_T1[1,1,:,:],kernel)
#         t23f = conv2d(np.real(T_T1[1,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[1,2,:,:]),kernel)

#         t31f = np.conj(t13f) 
#         t32f = np.conj(t23f) 
#         t33f = conv2d(T_T1[2,2,:,:],kernel)

#         T_T1 = np.array([[t11f, t12f, t13f], [t21f, t22f, t23f], [t31f, t32f, t33f]])

#     blue_ = norm_data(np.real(T_T1[0,0,:,:]))
#     red_ = norm_data(np.real(T_T1[1,1,:,:]))
#     green_ = norm_data(np.real(T_T1[2,2,:,:]))

#     rgb_uint8 = (np.dstack((red_,green_,blue_)) * 255) .astype(np.uint8)

#     return rgb_uint8[:,:,0],rgb_uint8[:,:,1],rgb_uint8[:,:,2]


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.utils.convert_matrices import T3_C3_mat

@time_it
def halphafp(infolder, outname=None, chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):

    if os.path.isfile(os.path.join(infolder,"T11.bin")):
        input_filepaths = [
        os.path.join(infolder,"T11.bin"),
        os.path.join(infolder,'T12_real.bin'), os.path.join(infolder,'T12_imag.bin'),  
        os.path.join(infolder,'T13_real.bin'), os.path.join(infolder,'T13_imag.bin'),
        os.path.join(infolder,"T22.bin"),
        os.path.join(infolder,'T23_real.bin'), os.path.join(infolder,'T23_imag.bin'),  
        os.path.join(infolder,"T33.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder,"C11.bin")):
        input_filepaths = [
        os.path.join(infolder,"C11.bin"),
        os.path.join(infolder,'C12_real.bin'), os.path.join(infolder,'C12_imag.bin'),  
        os.path.join(infolder,'C13_real.bin'), os.path.join(infolder,'C13_imag.bin'),
        os.path.join(infolder,"C22.bin"),
        os.path.join(infolder,'C23_real.bin'), os.path.join(infolder,'C23_imag.bin'),  
        os.path.join(infolder,"C33.bin"),
        ]

    else:
        print(f"Invalid C3 or T3 folder!!")

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "H_fp.tif"))
        output_filepaths.append(os.path.join(infolder, "alpha_fp.tif"))
        output_filepaths.append(os.path.join(infolder, "anisotropy_fp.tif"))
        output_filepaths.append(os.path.join(infolder, "e1_norm.tif"))
        output_filepaths.append(os.path.join(infolder, "e2_norm.tif"))
        output_filepaths.append(os.path.join(infolder, "e3_norm.tif"))
    
    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
            processing_func=process_chunk_halphafp,
            block_size=(512, 512), max_workers=max_workers, 
            num_outputs=6)

def process_chunk_halphafp(chunks, window_size, input_filepaths, *args):

    # additional_arg1 = args[0] if len(args) > 0 else None
    # additional_arg2 = args[1] if len(args) > 1 else None

    if 'T11' in input_filepaths[0] and 'T22' in input_filepaths[5] and 'T33' in input_filepaths[8]:
        t11_T1 = np.array(chunks[0])
        t12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
        t13_T1 = np.array(chunks[3])+1j*np.array(chunks[4])
        t21_T1 = np.conj(t12_T1)
        t22_T1 = np.array(chunks[5])
        t23_T1 = np.array(chunks[6])+1j*np.array(chunks[7])
        t31_T1 = np.conj(t13_T1)
        t32_T1 = np.conj(t23_T1)
        t33_T1 = np.array(chunks[8])

        T_T1 = np.array([[t11_T1, t12_T1, t13_T1], 
                     [t21_T1, t22_T1, t23_T1], 
                     [t31_T1, t32_T1, t33_T1]])
        # T_T1 = T3_C3_mat(T3)


    elif 'C11' in input_filepaths[0] and 'C22' in input_filepaths[5] and 'C33' in input_filepaths[8]:
        C11 = np.array(chunks[0])
        C12 = np.array(chunks[1])+1j*np.array(chunks[2])
        C13 = np.array(chunks[3])+1j*np.array(chunks[4])
        C21 = np.conj(C12)
        C22 = np.array(chunks[5])
        C23 = np.array(chunks[6])+1j*np.array(chunks[7])
        C31 = np.conj(C13)
        C32 = np.conj(C23)
        C33 = np.array(chunks[8])
        T_T1 = np.array([[C11, C12, C13], 
                         [C21, C22, C23], 
                         [C31, C32, C33]])

    else:
        raise ValueError("Invalid input matrices. Ensure the input is either T3 or C3 matrix foolder.")


    if window_size>1:
        kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)

        t11f = conv2d(T_T1[0,0,:,:],kernel)
        t12f = conv2d(np.real(T_T1[0,1,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,1,:,:]),kernel)
        t13f = conv2d(np.real(T_T1[0,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,2,:,:]),kernel)
        
        t21f = np.conj(t12f) 
        t22f = conv2d(T_T1[1,1,:,:],kernel)
        t23f = conv2d(np.real(T_T1[1,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[1,2,:,:]),kernel)

        t31f = np.conj(t13f) 
        t32f = np.conj(t23f) 
        t33f = conv2d(T_T1[2,2,:,:],kernel)

        T_T1 = np.array([[t11f, t12f, t13f], [t21f, t22f, t23f], [t31f, t32f, t33f]])

    _,_,rows,cols = np.shape(T_T1)
    T_T1 = T_T1.reshape(9, rows, cols)
    
    # Indices for vectorized access
    i, j = np.indices((rows, cols))
    
    T_T1 = np.dstack((T_T1[0,:,:],T_T1[1,:,:],T_T1[2,:,:],
                      T_T1[3,:,:],T_T1[4,:,:],T_T1[5,:,:],T_T1[6,:,:],T_T1[7,:,:],T_T1[8,:,:]))
    
    data = T_T1.reshape( T_T1.shape[0]*T_T1.shape[1], T_T1.shape[2]).reshape((-1,3,3))

    evals_, evecs_ = np.linalg.eig(data.reshape(-1, 3, 3))

    # Sort eigenvalues for each pixel in descending order; 
    sorted_indices = np.argsort(evals_, axis=-1)[:, ::-1] 
    
    # Reorder eigenvalues and eigenvectors based on sorted indices
    evals = np.take_along_axis(evals_, sorted_indices, axis=-1)  # Reorder eigenvalues
    
    # To reorder the eigenvectors, we use  indexing
    # Use `sorted_indices` to index along the second axis (the 3 components of the eigenvector)
    evecs = np.array([evecs_[i, :, sorted_indices[i]] for i in range(evecs_.shape[0])])
    
    # print('Eigen!')
    # Not sure if this is required; if enabled entropy values are different between polsarpro and this code
    # evals[:,0][evals[:,0] <0] = 0
    # evals[:,1][evals[:,1] <0] = 0
    # evals[:,2][evals[:,2] <0] = 0
    # evals[:,0][evals[:,0] >1] = 1
    # evals[:,1][evals[:,1] >1] = 1
    # evals[:,2][evals[:,2] >1] = 1
    
    eval_norm1 = (evals[:,0])/(evals[:,0] + evals[:,1]+ evals[:,2])
    eval_norm1[eval_norm1<0]=0
    # eval_norm1[eval_norm1>1]=1
    
    
    eval_norm2 = (evals[:,1])/(evals[:,0] + evals[:,1]+ evals[:,2])
    eval_norm2[eval_norm2<0]=0
    # eval_norm2[eval_norm2>1]=1
    
    
    eval_norm3 = (evals[:,2])/(evals[:,0] + evals[:,1]+evals[:,2])
    eval_norm3[eval_norm3<0]=0
    # eval_norm3[eval_norm3>1]=1
      
    
    # # %Alpha 1
    eig_vec_r1 = np.real(evecs[:,0,0])
    eig_vec_c1 = np.imag(evecs[:,0,0])
    alpha1 = np.arccos(np.sqrt(eig_vec_r1*eig_vec_r1 + eig_vec_c1*eig_vec_c1))*180/np.pi
    
    # # %Alpha 2
    eig_vec_r2 = np.real(evecs[:,0,1])
    eig_vec_c2 = np.imag(evecs[:,0,1])
    alpha2 = np.arccos(np.sqrt(eig_vec_r2*eig_vec_r2 + eig_vec_c2*eig_vec_c2))*180/np.pi
    
    # # %Alpha 3
    eig_vec_r3 = np.real(evecs[:,0,2])
    eig_vec_c3 = np.imag(evecs[:,0,2])
    alpha3 = np.arccos(np.sqrt(eig_vec_r3*eig_vec_r3 + eig_vec_c3*eig_vec_c3))*180/np.pi
        
    # # %Cloude Alpha
    alpha_ = (eval_norm1*alpha1 + eval_norm2*alpha2+ eval_norm3*alpha3)
    alpha_ = alpha_.reshape(rows,cols)
    
    # print('Alpha!')
    
    # # %Entropy
    H = - eval_norm1*np.log10(eval_norm1)/np.log10(3) - eval_norm2*np.log10(eval_norm2)/np.log10(3) - eval_norm3*np.log10(eval_norm3)/np.log10(3)
    H = H.reshape(rows,cols)

    # alpha1 = alpha1.reshape(rows,cols)
    # alpha2 = alpha2.reshape(rows,cols)
    # alpha3 = alpha3.reshape(rows,cols)
    
    ## POLARIMETRIC SCATTERING ANISOTROPY (A)
    Anisotropy = (eval_norm2-eval_norm3)/(eval_norm2+eval_norm3)
    
    return H,alpha_,Anisotropy.reshape(rows,cols),eval_norm1.reshape(rows,cols),eval_norm2.reshape(rows,cols),eval_norm3.reshape(rows,cols)


"""
A python package for processing Polarimetric Synthetic Aperture Radar (PolSAR) data.
"""
# polsartools/__init__.py

import warnings
warnings.filterwarnings("ignore")

__version__ = '0.5.1'  

# Import submodules
from . import polsar
from . import preprocessing
from . import utils 
from . import sensors
# Import functions from the submodules for direct access
from .polsar.fp import grvi, rvifp, mf3cf, mf4cf, dopfp, prvifp,nnedfp, neufp,halphafp
from .polsar.cp import cprvi, dopcp, misomega, mf3cc
from .polsar.dxp import dprvi, dopdp, prvidp, rvidp, halphadp
from .sensors.uavsar import uavsar_grd,uavsar_mlc
from .utils import convert_T3_C3,convert_C3_T3
from .preprocessing.filters import boxcar,rlee

## CPP functions
# import polsartools.refined_lee
# import polsartools.moving_average
# import polsartools.sum_arrays
# import polsartools.cprvicpp #process_chunk_cprvicpp
__all__ = [
    'uavsar_grd', 'uavsar_mlc', #import data from sensors
    # 'rl', 'boxcar', #import filters
    'convert_T3_C3', 'convert_C3_T3', 'pauliRGB',#import utils
    'grvi', 'rvifp', 'mf3cf', 'mf4cf', 'dopfp', 'prvifp', 'neufp', 'nnedfp', 'halphafp', # Full-pol
    'cprvi', 'dopcp', 'misomega', 'mf3cc',                 # Compact-pol
    'dprvi', 'dopdp', 'prvidp', 'rvidp', 'halphadp'        # Dual-cross-pol
    
]


import numpy as np
import glob
from osgeo import gdal,osr
import os
import sys
import simplekml
from polsartools.utils.utils import time_it
def write_bin_uav_old(file,wdata,lat,lon,dx,dy):
    
    [cols, rows] = wdata.shape

    driver = gdal.GetDriverByName("ENVI")
    outdata = driver.Create(file, rows, cols, 1, gdal.GDT_Float32)
    outdata.SetGeoTransform([lon,dx,0,lat,0,dy])
    outdata.SetProjection("EPSG:4326")
    outdata.SetDescription(file)
    outdata.GetRasterBand(1).WriteArray(wdata)
    # outdata.GetRasterBand(1).SetNoDataValue(np.NaN)##if you want these values transparent
    outdata.FlushCache() ##saves to disk!!

def write_bin_uav(file, wdata, lat, lon, dx, dy, sensor_type="UAVSAR"):
    [rows, cols] = wdata.shape
    driver = gdal.GetDriverByName("ENVI")
    outdata = driver.Create(file, cols, rows, 1, gdal.GDT_Float32)
    outdata.SetGeoTransform([lon, dx, 0, lat, 0, dy])
    outdata.SetProjection("EPSG:4326")
    outdata.GetRasterBand(1).WriteArray(wdata)
    outdata.FlushCache()
    # if dy<0:
    #     dy = -dy
    # Write the header file
    # header_filename = file.replace('.bin', '.hdr') 
    # with open(header_filename, 'w') as header_file:
    #     header_file.write("ENVI\n")
    #     header_file.write(f"description = {{{file}}}\n")
    #     header_file.write(f"samples = {cols}\n")
    #     header_file.write(f"lines = {rows}\n")
    #     header_file.write("bands = 1\n")
    #     header_file.write("header offset = 0\n")
    #     header_file.write("file type = ENVI Standard\n")
    #     header_file.write("data type = 4\n")  # Float32
    #     header_file.write("interleave = bsq\n")
    #     # header_file.write("wavelength units = METERS\n")
    #     header_file.write("byte order = 0\n")
    #     header_file.write(f"sensor type = {sensor_type}\n")
    #     map_info = f"map info = {{Geographic Lat/Lon, 1, 1, {lon}, {lat}, {dx}, {dy}, WGS-84}}\n"
    #     header_file.write(map_info)
    #     # Write the projection information
    #     # header_file.write("projection = GEOGCS[\"WGS 84\",DATUM[\"WGS_1984\",SPHEROID[\"WGS_84\",6378137,298.257223563]],PRIMEM[\"Greenwich\",0],UNIT[\"degree\",0.017453292519943295],AUTHORITY[\"EPSG\",\"4326\"]]\n")
    #     header_file.write(f"band names = {{{os.path.basename(file)}}}\n")

def create_kml_polygon(corner_coords, output_filename):
    
    kml = simplekml.Kml()
    pol = kml.newpolygon(name="Polygon")
    polygon_coords = corner_coords + [corner_coords[0]]
    pol.outerboundaryis.coords = polygon_coords
    pol.style.polystyle.color = simplekml.Color.changealphaint(150, simplekml.Color.red)  
    kml.save(output_filename)

def create_extent(annFile):
    inFolder = os.path.dirname(annFile)
    annFile = open(annFile, 'r')
    for line in annFile:
        if "Approximate Upper Left Latitude" in line:
            uly = float(line.split('=')[1].split(';')[0])
        if "Approximate Upper Left Longitude" in line:
            ulx = float(line.split('=')[1].split(';')[0])
        if "Approximate Lower Right Latitude" in line:
            lry = float(line.split('=')[1].split(';')[0])
        if "Approximate Lower Right Longitude" in line:
            lrx = float(line.split('=')[1].split(';')[0])
        if "Approximate Lower Left Latitude" in line:
            lly = float(line.split('=')[1].split(';')[0])
        if "Approximate Lower Left Longitude" in line:
            llx = float(line.split('=')[1].split(';')[0]) 
        if "Approximate Upper Right Longitude" in line:
            urx = float(line.split('=')[1].split(';')[0])
        if "Approximate Upper Right Latitude" in line:
            ury = float(line.split('=')[1].split(';')[0]) 

    corner_coordinates = [
        (ulx, uly),  # (longitude, latitude) coordinate 1
        (urx, ury),  # coordinate 2
        (lrx, lry),  # coordinate 3
        (llx, lly),  # coordinate 4
    ]
    # print(corner_coordinates)

    output_file = os.path.join(inFolder,"scene_extent.kml")
    create_kml_polygon(corner_coordinates, output_file)
    
@time_it    
def uavsar_grd(annFile):
    inFolder = os.path.dirname(annFile)
    create_extent(annFile)
    annFile = open(annFile, 'r')
    for line in annFile:
        if "grd_mag.set_rows" in line:
            rows = int(line.split('=')[1].split(';')[0])
        if "grd_mag.set_cols" in line:
            cols = int(line.split('=')[1].split(';')[0])
        if "grd_mag.row_addr" in line:
            lat = float(line.split('=')[1].split(';')[0])
        if "grd_mag.col_addr" in line:
            lon = float(line.split('=')[1].split(';')[0])
        if "grd_mag.row_mult" in line and "(deg/pixel)" in line:
            dy = float(line.split('=')[1].split(';')[0])
        if "grd_mag.col_mult" in line and "(deg/pixel)" in line:
            dx = float(line.split('=')[1].split(';')[0])

    outFolder = inFolder+'/C3'
    if not os.path.isdir(outFolder):
        print("C3 folder does not exist. \nCreating folder {}".format(outFolder))
        os.mkdir(outFolder)
    else:
        print("C3 folder exists. \nReplacing C3 elements in folder {}".format(outFolder))

    hhhh = np.fromfile(glob.glob(inFolder+'/*HHHH*.grd')[0], dtype='<f',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C11.bin',hhhh,lat,lon,dx,dy)
    del hhhh
    vvvv = np.fromfile(glob.glob(inFolder+'/*VVVV*.grd')[0], dtype='<f',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C33.bin',vvvv,lat,lon,dx,dy)
    del vvvv
    hvhv = np.fromfile(glob.glob(inFolder+'/*HVHV*.grd')[0], dtype='<f',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C22.bin',2*hvhv,lat,lon,dx,dy)
    del hvhv
    hhhv = np.fromfile(glob.glob(inFolder+'/*HHHV*.grd')[0], dtype='<F',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C12_real.bin',np.real(np.sqrt(2)*hhhv),lat,lon,dx,dy)
    write_bin_uav(outFolder+'/C12_imag.bin',np.imag(np.sqrt(2)*hhhv),lat,lon,dx,dy)
    del hhhv
    hhvv = np.fromfile(glob.glob(inFolder+'/*HHVV*.grd')[0], dtype='<F',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C13_real.bin',np.real(hhvv),lat,lon,dx,dy)
    write_bin_uav(outFolder+'/C13_imag.bin',np.imag(hhvv),lat,lon,dx,dy)
    del hhvv
    hvvv = np.fromfile(glob.glob(inFolder+'/*HVVV*.grd')[0], dtype='<F',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C23_real.bin',np.real(np.sqrt(2)*hvvv),lat,lon,dx,dy)
    write_bin_uav(outFolder+'/C23_imag.bin',np.imag(np.sqrt(2)*hvvv),lat,lon,dx,dy)
    del hvvv

    file = open(outFolder +'/config.txt',"w+")
    file.write('Nrow\n%d\n---------\nNcol\n%d\n---------\nPolarCase\nmonostatic\n---------\nPolarType\nfull'%(rows,cols))
    file.close()  
    print("Extracted C3 files to %s"%outFolder)
@time_it  
def uavsar_mlc(annFile):
    create_extent(annFile)
    inFolder = os.path.dirname(annFile)
    create_extent(annFile)
    annFile = open(annFile, 'r')
    for line in annFile:
        if "mlc_mag.set_rows" in line:
            rows = int(line.split('=')[1].split(';')[0])
        if "mlc_mag.set_cols" in line:
            cols = int(line.split('=')[1].split(';')[0])
        if "grd_mag.row_addr" in line:
            lat = float(line.split('=')[1].split(';')[0])
        if "grd_mag.col_addr" in line:
            lon = float(line.split('=')[1].split(';')[0])
        if "grd_mag.row_mult" in line and "(deg/pixel)" in line:
            dy = float(line.split('=')[1].split(';')[0])
        if "grd_mag.col_mult" in line and "(deg/pixel)" in line:
            dx = float(line.split('=')[1].split(';')[0])
        # if "set_plon" in line  and "(deg)" in line:
        #     lon = float(line.split('=')[1].split(';')[0])
        # if "set_plat" in line  and "(deg)" in line:
        #     lat = float(line.split('=')[1].split(';')[0])
        # if "mlc_mag.row_mult" in line and "(m/pixel) " in line:
        #     dy = float(line.split('=')[1].split(';')[0])*0.00001
        # if "mlc_mag.col_mult" in line and "(m/pixel) " in line:
        #     dx = float(line.split('=')[1].split(';')[0])*0.00001

    outFolder = inFolder+'/C3'
    if not os.path.isdir(outFolder):
        print("C3 folder does not exist. \nCreating folder {}".format(outFolder))
        os.mkdir(outFolder)
    else:
        print("C3 folder exists. \nReplacing C3 elements in folder {}".format(outFolder))

        
    hhhh = np.fromfile(glob.glob(inFolder+'/*HHHH*.mlc')[0], dtype='<f',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C11.bin',hhhh,lat,lon,dx,dy)

    del hhhh
    vvvv = np.fromfile(glob.glob(inFolder+'/*VVVV*.mlc')[0], dtype='<f',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C33.bin',vvvv,lat,lon,dx,dy)
    del vvvv
    hvhv = np.fromfile(glob.glob(inFolder+'/*HVHV*.mlc')[0], dtype='<f',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C22.bin',2*hvhv,lat,lon,dx,dy)
    del hvhv
    hhhv = np.fromfile(glob.glob(inFolder+'/*HHHV*.mlc')[0], dtype='<F',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C12_real.bin',np.real(np.sqrt(2)*hhhv),lat,lon,dx,dy)
    write_bin_uav(outFolder+'/C12_imag.bin',np.imag(np.sqrt(2)*hhhv),lat,lon,dx,dy)
    del hhhv
    hhvv = np.fromfile(glob.glob(inFolder+'/*HHVV*.mlc')[0], dtype='<F',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C13_real.bin',np.real(hhvv),lat,lon,dx,dy)
    write_bin_uav(outFolder+'/C13_imag.bin',np.imag(hhvv),lat,lon,dx,dy)
    del hhvv
    hvvv = np.fromfile(glob.glob(inFolder+'/*HVVV*.mlc')[0], dtype='<F',).reshape(rows,cols)
    write_bin_uav(outFolder+'/C23_real.bin',np.real(np.sqrt(2)*hvvv),lat,lon,dx,dy)
    write_bin_uav(outFolder+'/C23_imag.bin',np.imag(np.sqrt(2)*hvvv),lat,lon,dx,dy)
    del hvvv

    file = open(outFolder +'/config.txt',"w+")
    file.write('Nrow\n%d\n---------\nNcol\n%d\n---------\nPolarCase\nmonostatic\n---------\nPolarType\nfull'%(rows,cols))
    file.close()   
    print("Extracted C3 files to %s"%outFolder)



import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.utils.convert_matrices import C3_T3_mat

@time_it
def rvifp(infolder, outname=None,  chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):

    if os.path.isfile(os.path.join(infolder,"T11.bin")):
        input_filepaths = [
        os.path.join(infolder,"T11.bin"),
        
        os.path.join(infolder,'T12_real.bin'), os.path.join(infolder,'T12_imag.bin'),  
        os.path.join(infolder,'T13_real.bin'), os.path.join(infolder,'T13_imag.bin'),
        os.path.join(infolder,"T22.bin"),
        os.path.join(infolder,'T23_real.bin'), os.path.join(infolder,'T23_imag.bin'),  
     
        os.path.join(infolder,"T33.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder,"C11.bin")):

        input_filepaths = [
        os.path.join(infolder,"C11.bin"),
        
        os.path.join(infolder,'C12_real.bin'), os.path.join(infolder,'C12_imag.bin'),  
        os.path.join(infolder,'C13_real.bin'), os.path.join(infolder,'C13_imag.bin'),
        os.path.join(infolder,"C22.bin"),
        os.path.join(infolder,'C23_real.bin'), os.path.join(infolder,'C23_imag.bin'),  
     
        os.path.join(infolder,"C33.bin"),
        ]

    else:
        print(f"Invalid C3 or T3 folder!!")
    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "rvifp.tif"))
    
    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
            processing_func=process_chunk_rvifp,block_size=(512, 512), max_workers=max_workers,  num_outputs=1)


def process_chunk_rvifp(chunks, window_size,input_filepaths,*args):

    t11_T1 = np.array(chunks[0])
    t12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
    t13_T1 = np.array(chunks[3])+1j*np.array(chunks[4])
    t21_T1 = np.conj(t12_T1)
    t22_T1 = np.array(chunks[5])
    t23_T1 = np.array(chunks[6])+1j*np.array(chunks[7])
    t31_T1 = np.conj(t13_T1)
    t32_T1 = np.conj(t23_T1)
    t33_T1 = np.array(chunks[8])

    T_T1 = np.array([[t11_T1, t12_T1, t13_T1], 
                     [t21_T1, t22_T1, t23_T1], 
                     [t31_T1, t32_T1, t33_T1]])

    if window_size>1:
        kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)

        t11f = conv2d(T_T1[0,0,:,:],kernel)
        t12f = conv2d(np.real(T_T1[0,1,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,1,:,:]),kernel)
        t13f = conv2d(np.real(T_T1[0,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,2,:,:]),kernel)
        
        t21f = np.conj(t12f) 
        t22f = conv2d(T_T1[1,1,:,:],kernel)
        t23f = conv2d(np.real(T_T1[1,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[1,2,:,:]),kernel)

        t31f = np.conj(t13f) 
        t32f = np.conj(t23f) 
        t33f = conv2d(T_T1[2,2,:,:],kernel)

        T_T1 = np.array([[t11f, t12f, t13f], [t21f, t22f, t23f], [t31f, t32f, t33f]])


    reshaped_arr = T_T1.reshape(3, 3, -1).transpose(2, 0, 1)
    eigenvalues = np.linalg.eigvals(reshaped_arr)
    sorted_eigenvalues = np.sort(eigenvalues, axis=1)[:, ::-1]
    sorted_eigenvalues = sorted_eigenvalues.reshape(T_T1.shape[2], T_T1.shape[3], 3)
    
    p1 = sorted_eigenvalues[:,:,0]/(sorted_eigenvalues[:,:,0] + sorted_eigenvalues[:,:,1] + sorted_eigenvalues[:,:,2])
    p2 = sorted_eigenvalues[:,:,1]/(sorted_eigenvalues[:,:,0] + sorted_eigenvalues[:,:,1] + sorted_eigenvalues[:,:,2])
    p3 = sorted_eigenvalues[:,:,2]/(sorted_eigenvalues[:,:,0] + sorted_eigenvalues[:,:,1] + sorted_eigenvalues[:,:,2])
    
    p1[p1<0] = 0
    p2[p2<0] = 0
    p3[p3<0] = 0
        
    p1[p1>1] = 1
    p2[p2>1] = 1
    p3[p3>1] = 1

    
    rvi = np.real((4*p3)/(p1 + p2 + p3))

    idx = np.argwhere(rvi>1)

    rvi[idx] = (3/4)*rvi[idx]
    rvi[~idx] = rvi[~idx]
    rvi[rvi==0] = np.nan

    rvi = np.real(rvi)

    return rvi



import os
import pytest
import numpy as np
import rasterio
from osgeo import gdal
from polsartools.polsar.fullpol.processing import process_raster_in_chunks

@pytest.fixture
def create_dummy_geotiff():
    """
    Fixture to create a dummy single-band GeoTIFF file for testing.
    """
    directory = 'test_data'
    filename = 'test_raster.tif'
    filepath = os.path.join(directory, filename)
    
    if not os.path.exists(directory):
        os.makedirs(directory)
    
    # Define data type mapping
    dtype = 'float32'
    dtype_map = {
        'float32': rasterio.float32,
        'int16': rasterio.int16,
        'uint16': rasterio.uint16,
        # Add more mappings if needed
    }
    
    if dtype not in dtype_map:
        raise ValueError(f"Unsupported dtype: {dtype}")

    # Create a dummy GeoTIFF file
    with rasterio.open(filepath, 'w', driver='GTiff', height=512, width=512, count=1, dtype=dtype_map[dtype], crs='EPSG:4326', transform=rasterio.Affine(1, 0, 0, 0, -1, 0)) as dst:
        data = np.random.random((512, 512)).astype(dtype)
        dst.write(data, 1)
    
    yield filepath

    # Clean up
    os.remove(filepath)
    os.rmdir(directory)

def test_process_raster_in_chunks(create_dummy_geotiff):
    """
    Test processing of raster file in chunks.
    """
    input_file = create_dummy_geotiff
    output_file = os.path.join('test_data', 'processed_raster.tif')
    param1 = 2  # Example parameter for processing
    param2 = 5  # Example parameter for processing
    chunk_size = (256, 256)  # Smaller chunks for testing

    process_raster_in_chunks(input_file, output_file, param1, param2, chunk_size)

    # Verify the output file
    with rasterio.open(output_file) as dst:
        assert dst is not None, "Output file could not be opened"
        assert dst.count == 1  # Verify single-band
        assert dst.width == 512
        assert dst.height == 512
        assert dst.dtypes[0] == 'float32'  # Ensure dtype matches

    # Clean up
    os.remove(output_file)



import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d, eig22

@time_it
def rvidp(infolder, outname=None,  chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):
    input_filepaths = [
        os.path.join(infolder, "C11.bin"), 
        os.path.join(infolder, "C12_real.bin"),
        os.path.join(infolder, "C12_imag.bin"),
        os.path.join(infolder, "C22.bin")
    ]
    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "rvidp.tif"))

    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,processing_func=process_chunk_rvidp,block_size=(512, 512), max_workers=max_workers,  num_outputs=1)

def process_chunk_rvidp(chunks, window_size,*args):
    kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)
    c11_T1 = np.array(chunks[0])
    c12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
    c21_T1 = np.conj(c12_T1)
    c22_T1 = np.array(chunks[3])

    if window_size>1:
        c11_T1r = conv2d(np.real(c11_T1),kernel)
        c11_T1i = conv2d(np.imag(c11_T1),kernel)
        c11s = c11_T1r+1j*c11_T1i

        c12_T1r = conv2d(np.real(c12_T1),kernel)
        c12_T1i = conv2d(np.imag(c12_T1),kernel)
        c12s = c12_T1r+1j*c12_T1i


        c21_T1r = conv2d(np.real(c21_T1),kernel)
        c21_T1i = conv2d(np.imag(c21_T1),kernel)
        c21s = c21_T1r+1j*c21_T1i


        c22_T1r = conv2d(np.real(c22_T1),kernel)
        c22_T1i = conv2d(np.imag(c22_T1),kernel)
        c22s = c22_T1r+1j*c22_T1i

        
        c2_det = (c11s*c22s-c12s*c21s)
        c2_trace = c11s+c22s
        rvidp = np.real(4*c22s/c2_trace)
    else:
        c2_det = (c11_T1*c22_T1-c12_T1*c21_T1)
        c2_trace = c11_T1+c22_T1
        rvidp = np.real(4*c22_T1/c2_trace)

    return rvidp


import os, tempfile
from osgeo import gdal
import numpy as np
from functools import wraps
from concurrent.futures import ProcessPoolExecutor,as_completed
from tqdm import tqdm
import time

def time_it(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        try:
            result = func(*args, **kwargs)  # Call the original function
            end_time = time.time()
            print(f"Execution time for {func.__name__}: {end_time - start_time:.2f} seconds")
            return result
        except Exception as e:
            # If an exception occurs, re-raise it but don't print the execution time
            raise e
    return wrapper

def conv2d(a, f):
    filt = np.zeros(a.shape)
    wspad = int(f.shape[0]/2)
    s = f.shape + tuple(np.subtract(a.shape, f.shape) + 1)
    strd = np.lib.stride_tricks.as_strided
    subM = strd(a, shape=s, strides=a.strides * 2)
    filt_data = np.einsum('ij,ijkl->kl', f, subM)
    filt[wspad:wspad+filt_data.shape[0], wspad:wspad+filt_data.shape[1]] = filt_data
    return filt

def eig22(c2):
    c11 = c2[:, :, 0].flatten()
    c12 = c2[:, :, 1].flatten()
    c21 = c2[:, :, 2].flatten()
    c22 = c2[:, :, 3].flatten()
    trace = -(c11 + c22)
    det = c11 * c22 - c12 * c21
    sqdiscr = np.sqrt(trace * trace - 4 * det)
    lambda1 = -(trace + sqdiscr) * 0.5
    lambda2 = -(trace - sqdiscr) * 0.5
    return lambda1, lambda2
 

def process_chunks_parallel(input_filepaths, output_filepaths, 
                            window_size, write_flag, processing_func, 
                            block_size=(512, 512), max_workers=None, 
                            num_outputs=1, chi_in=None,psi_in=None):


    if len(input_filepaths) not in [2, 4, 9]:
        raise ValueError("This function only supports 2, 4, or 9 input rasters.")

    if max_workers is None:
        max_workers = os.cpu_count()-1  # Use all available CPUs
        # max_workers = 1
    input_datasets = [gdal.Open(fp, gdal.GA_ReadOnly) for fp in input_filepaths]
    if any(ds is None for ds in input_datasets):
        raise FileNotFoundError("One or more input files could not be opened.")

    raster_width = input_datasets[0].RasterXSize
    raster_height = input_datasets[0].RasterYSize
    geotransform = input_datasets[0].GetGeoTransform()
    projection = input_datasets[0].GetProjection()

    # Ensure block_size does not exceed raster dimensions
    adjusted_block_size_x = min(block_size[0], raster_width)
    adjusted_block_size_y = min(block_size[1], raster_height)

    merged_arrays = [np.zeros((raster_height, raster_width), dtype=np.float32) for _ in range(num_outputs)] if not write_flag else None

    tasks = []

    """ with progress bar"""
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
    # with ThreadPoolExecutor(max_workers=max_workers) as executor:
        tasks = []
        for y in range(0, raster_height, adjusted_block_size_y):
            for x in range(0, raster_width, adjusted_block_size_x):
                read_block_width = min(adjusted_block_size_x, raster_width - x)
                read_block_height = min(adjusted_block_size_y, raster_height - y)
                args_ = (input_filepaths, x, y, read_block_width, read_block_height, window_size, raster_width, raster_height, chi_in, psi_in)
                # print(f"Submitting task for chunk at ({x}, {y})")
                tasks.append(executor.submit(process_and_write_chunk, args_, processing_func, num_outputs))

        # Initialize tqdm progress bar with the total number of tasks
        with tqdm(total=len(tasks), desc=f"Progress ", unit=" block") as pbar:
            temp_files = []
            for future in as_completed(tasks):
                try:
                    result = future.result()
                    if result is None:
                        raise ValueError("Block processing returned None")
                    temp_paths, x_start, y_start = result
                    if write_flag:
                        temp_files.append((temp_paths, x_start, y_start))
                    else:
                        for i in range(num_outputs):
                            temp_dataset = gdal.Open(temp_paths[i], gdal.GA_ReadOnly)
                            temp_band = temp_dataset.GetRasterBand(1)
                            temp_chunk = temp_band.ReadAsArray()
                            print(type(temp_chunk))
                            temp_height, temp_width = temp_chunk.shape
                            merged_arrays[i][y_start:y_start + temp_height, x_start:x_start + temp_width] = temp_chunk
                            temp_dataset = None
                            os.remove(temp_paths[i])
                    
                    pbar.update(1)
                except Exception as e:
                    print(f"Error in processing task: {e}")


    if write_flag:
        merge_temp_files(output_filepaths, temp_files, raster_width, raster_height, geotransform, projection, num_outputs)
        for temp_path_set, _, _ in temp_files:
            for temp_path in temp_path_set:
                os.remove(temp_path)
    else:
        return merged_arrays

def read_chunk_with_overlap(filepath, x_start, y_start, width, height, window_size):
    dataset = gdal.Open(filepath, gdal.GA_ReadOnly)
    if dataset is None:
        raise FileNotFoundError(f"Cannot open {filepath}")

    band = dataset.GetRasterBand(1)
    # Handle edge cases for window overlap near the borders of the raster
    xoff = max(x_start - window_size, 0)
    yoff = max(y_start - window_size, 0)
    
    xsize = min(width + 2 * window_size, dataset.RasterXSize - xoff)
    ysize = min(height + 2 * window_size, dataset.RasterYSize - yoff)


    # Check if the raster size is less than or equal to  block size read entire raster
    if width >= dataset.RasterXSize and height >= dataset.RasterYSize:
        chunk = band.ReadAsArray(xoff=xoff, yoff=yoff, win_xsize=xsize, win_ysize=ysize)
        dataset = None  
        return chunk


    # Determine if the chunk is near the borders of the dataset
    is_left_border = x_start == 0
    is_right_border = x_start + width >= dataset.RasterXSize
    is_top_border = y_start == 0
    is_bottom_border = y_start + height >= dataset.RasterYSize

    # Handle corner cases
    if is_left_border and is_top_border:
        # Top-left corner: no overlap on top and left
        xoff = x_start
        yoff = y_start
        xsize = min(width + window_size, dataset.RasterXSize - xoff)
        ysize = min(height + window_size, dataset.RasterYSize - yoff)
    elif is_left_border and is_bottom_border:
        # Bottom-left corner: no overlap on bottom and left
        xoff = x_start
        yoff = max(y_start - window_size, 0)
        xsize = min(width + window_size, dataset.RasterXSize - xoff)
        ysize = min(height + window_size, dataset.RasterYSize - yoff)
    elif is_right_border and is_top_border:
        # Top-right corner: no overlap on top and right
        xoff = max(x_start - window_size, 0)
        yoff = y_start
        xsize = min(width + window_size, dataset.RasterXSize - xoff)
        ysize = min(height + window_size, dataset.RasterYSize - yoff)
    elif is_right_border and is_bottom_border:
        # Bottom-right corner: no overlap on bottom and right
        xoff = max(x_start - window_size, 0)
        yoff = max(y_start - window_size, 0)
        xsize = min(width + window_size, dataset.RasterXSize - xoff)
        ysize = min(height + window_size, dataset.RasterYSize - yoff)
    elif is_left_border:
        # Left edge: no overlap on left, adjust only horizontally
        xoff = x_start
        xsize = min(width + window_size, dataset.RasterXSize - xoff)
    elif is_right_border:
        # Right edge: no overlap on right, adjust only horizontally
        xsize = min(width + window_size, dataset.RasterXSize - x_start)
    elif is_top_border:
        # Top edge: no overlap on top, adjust only vertically
        yoff = y_start
        ysize = min(height + window_size, dataset.RasterYSize - yoff)
    elif is_bottom_border:
        # Bottom edge: no overlap on bottom, adjust only vertically
        ysize = min(height + window_size, dataset.RasterYSize - y_start)

    # Read the adjusted chunk with overlap
    chunk = band.ReadAsArray(xoff=xoff, yoff=yoff, win_xsize=xsize, win_ysize=ysize)

    dataset = None  # Close the dataset after reading
    return chunk

def write_chunk_to_temp_file(processed_chunks, x_start, y_start, block_width, block_height, window_size, raster_width, raster_height, num_outputs):
    temp_paths = []
    
    for i in range(num_outputs):
        temp_fd, temp_path = tempfile.mkstemp(suffix=f'_output_{i}.tif')
        os.close(temp_fd)

        driver = gdal.GetDriverByName('GTiff')
        temp_dataset = driver.Create(temp_path, block_width, block_height, 1, gdal.GDT_Float32)
        geotransform = (x_start, 1, 0, y_start, 0, -1)
        temp_dataset.SetGeoTransform(geotransform)
        temp_dataset.SetProjection('')

        temp_band = temp_dataset.GetRasterBand(1)

        # Check if the block size is equal to the raster size
        if block_width >= raster_width and block_height >= raster_height:
            # temp_band.WriteArray(processed_chunks[i][:-window_size//2, :][:, :-window_size//2])  
            
            # rows, cols = processed_chunks[i].shape
            
            # # Calculate the shift
            # shift = window_size // 2
            # # Create a new array with dimensions adjusted for the shift
            # new_rows = rows + shift
            # new_cols = cols + shift
            # new_arr = np.zeros((new_rows, new_cols), dtype=processed_chunks[i].dtype)
            # new_arr[shift:shift+rows, shift:shift+cols] = processed_chunks[i]
            # new_arr[:shift, :shift] = processed_chunks[i][-shift:, -shift:]

            # temp_band.WriteArray(new_arr)       
            temp_band.WriteArray(processed_chunks[i]) 
            temp_dataset.FlushCache()
            temp_dataset = None
            temp_paths.append(temp_path)
            continue  # Skip to the next output
        
        # Determine the boundaries
        is_left_border = x_start == 0
        is_right_border = x_start + block_width >= raster_width
        is_top_border = y_start == 0
        is_bottom_border = y_start + block_height >= raster_height

        if is_left_border and is_top_border:
            # Top-left corner: reduce only from right and bottom
            temp_band.WriteArray(processed_chunks[i][:-window_size, :][:, :-window_size])
        elif is_left_border and is_bottom_border:
            # Bottom-left corner: reduce only from right and top
            temp_band.WriteArray(processed_chunks[i][window_size:, :][:, :-window_size])
        elif is_right_border and is_top_border:
            # Top-right corner: reduce only from left and bottom
            temp_band.WriteArray(processed_chunks[i][:-window_size, :][:, window_size:])
        elif is_right_border and is_bottom_border:
            # Bottom-right corner: reduce only from left and top
            temp_band.WriteArray(processed_chunks[i][window_size:, :][:, window_size:])
        elif is_left_border:
            # Left edge: reduce from right, top, and bottom
            temp_band.WriteArray(processed_chunks[i][window_size:-window_size,:-window_size ])  # Skip left
        elif is_right_border:
            # Right edge: reduce from left, top, and bottom
            temp_band.WriteArray(processed_chunks[i][ window_size:-window_size,window_size:])  # Skip right
        elif is_top_border:
            # Top edge: reduce from left, right, and bottom
            temp_band.WriteArray(processed_chunks[i][ :-window_size,window_size:-window_size])  # Skip top
        elif is_bottom_border:
            # Bottom edge: reduce from left, right, and top
            temp_band.WriteArray(processed_chunks[i][window_size:,window_size:-window_size])  # Skip bottom
        else:
            # Non-border chunks: apply window size reduction
            temp_band.WriteArray(processed_chunks[i][window_size:-window_size, window_size:-window_size])

        temp_dataset.FlushCache()
        temp_dataset = None

        temp_paths.append(temp_path)

    return temp_paths, x_start, y_start

def merge_temp_files(output_filepaths, temp_files, raster_width, raster_height, geotransform, projection, num_outputs):
    for i in range(num_outputs):
        if '.tif' in output_filepaths[0]:
            driver = gdal.GetDriverByName('GTiff')
            output_dataset = driver.Create(output_filepaths[i], raster_width, raster_height, 1, gdal.GDT_Float32,
                                       options=['COMPRESS=LZW','BIGTIFF=YES','TILED=YES'])
        
        if '.bin' in output_filepaths[0]:
            driver = gdal.GetDriverByName('ENVI')
            output_dataset = driver.Create(output_filepaths[i], raster_width, raster_height, 1, gdal.GDT_Float32,
                                       )

        output_dataset.SetGeoTransform(geotransform)
        output_dataset.SetProjection(projection)
        output_band = output_dataset.GetRasterBand(1)

        for temp_path_set, x_start, y_start in temp_files:
            temp_dataset = gdal.Open(temp_path_set[i], gdal.GA_ReadOnly)
            temp_band = temp_dataset.GetRasterBand(1)

            temp_chunk = temp_band.ReadAsArray()
            temp_height, temp_width = temp_chunk.shape
            
            if len(temp_files) == 1:
                x_start, y_start = 0, 0
                
            output_band.WriteArray(temp_chunk, xoff=x_start, yoff=y_start)
            # print(x_start,y_start)
            temp_dataset = None

        output_dataset.FlushCache()
        output_dataset = None
        print(f"Saved file {output_filepaths[i]}")
def process_and_write_chunk(args, processing_func, num_outputs):
    try:
        (input_filepaths, x_start, y_start, read_block_width, read_block_height, window_size, raster_width, raster_height, chi_in,psi_in) = args

        chunks = [read_chunk_with_overlap(fp, x_start, y_start, read_block_width, read_block_height, window_size) for fp in input_filepaths]
        processed_chunks = processing_func(chunks, window_size, input_filepaths, chi_in,psi_in)

        if num_outputs == 1:
            processed_chunks = [processed_chunks]

        temp_paths, temp_x_start, temp_y_start = write_chunk_to_temp_file(
            processed_chunks,
            x_start, y_start, read_block_width, read_block_height, window_size, raster_width, raster_height, num_outputs
        )

        return temp_paths, temp_x_start, temp_y_start
    except Exception as e:
        print(f"Error in processing chunk: {e}")
        return None



import numpy as np
# import process_arrays,refined_lee  # The compiled module
import polsartools
from functools import wraps
from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm import tqdm
import time

def time_it(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        try:
            result = func(*args, **kwargs)  # Call the original function
            end_time = time.time()
            print(f"Execution time for {func.__name__}: {end_time - start_time:.2f} seconds")
            return result
        except Exception as e:
            # If an exception occurs, re-raise it but don't print the execution time
            raise e
    return wrapper
def process_chunk_cprvi(chunks, window_size,input_filepaths,chi_in,psi_in):

    kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)
    c11_T1 = np.array(chunks[0])
    c12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
    c21_T1 = np.conj(c12_T1)
    c22_T1 = np.array(chunks[3])

    ncols,nrows = np.shape(c11_T1)
    
    fp22 = np.zeros((ncols,nrows))
    l_lambda = np.zeros((ncols,nrows))

    wsi=wsj=window_size

    inci=int(np.fix(wsi/2)) # Up & down movement margin from the central row
    incj=int(np.fix(wsj/2)) # Left & right movement from the central column
    # % Starting row and column fixed by the size of the patch extracted from the image of 21/10/1999

    starti=int(np.fix(wsi/2)) # Starting row for window processing
    startj=int(np.fix(wsj/2)) # Starting column for window processing

    stopi= int(nrows-inci)-1 # Stop row for window processing
    stopj= int(ncols-incj)-1 # Stop column for window processing

    for ii in np.arange(startj,stopj+1):

        for jj in np.arange(starti,stopi+1):
            
            C11c = np.nanmean(c11_T1[ii-inci:ii+inci+1,jj-incj:jj+incj+1])#i sample
            C12c = np.nanmean(c12_T1[ii-inci:ii+inci+1,jj-incj:jj+incj+1])#i sample
            C21c = np.nanmean(c21_T1[ii-inci:ii+inci+1,jj-incj:jj+incj+1])#i sample
            C22c = np.nanmean(c22_T1[ii-inci:ii+inci+1,jj-incj:jj+incj+1])#i sample

            C0 = np.array([[C11c,C12c], [C21c, C22c]]);
            
            # %% GD_VI -- VV-VH/VV-HH
            if np.isnan(np.real(C0)).any() or np.isinf(np.real(C0)).any() or np.isneginf(np.real(C0)).any():
                C0 = np.array([[0,0],[0,0]])

            # Stokes Parameter
            s0 = C11c + C22c;
            s1 = C11c - C22c;
            s2 = (C12c + C21c);

            if (chi_in >= 0):
                s3 = (1j*(C12c - C21c)); # The sign is according to RC or LC sign !!
            if (chi_in < 0):
                s3 = -(1j*(C12c - C21c)); # The sign is according to RC or LC sign !!
            
            k11 = s0
            k12 = 0
            k13 = s2
            k14 = 0
            k21 = k12 
            k22 = 0
            k23 = 0
            k24 = s1
            k31 = k13
            k32 = k23
            k33 = 0 
            k34 = 0;
            k41 = k14; 
            k42 = k24; 
            k43 = k34; 
            k44 = s3;

            K_T = 0.5*np.array([[k11,k12,k13,k14], [k21,k22,k23,k24], 
                [k31, k32, k33, k34], [k41,k42,k43,k44]])       

            # Stokes vector child products
            SC = ((s0)-(s3))/2;
            OC = ((s0)+(s3))/2;
            
            min_sc_oc = min(SC,OC);
            max_sc_oc = max(SC,OC);                
            
            K_depol = np.array([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]);
            
            # GD_DEPOL
            
            num1 = np.matmul(K_T.T,K_depol);
            num = np.trace(num1);
            den1 = np.sqrt(abs(np.trace(np.matmul(K_T.T,K_T))));
            den2 = np.sqrt(abs(np.trace(np.matmul(K_depol.T,K_depol))));
            den = den1*den2;
            
            temp_aa = np.real(2*np.arccos(num/den)*180/np.pi);
            GD_t1_depol = np.real(temp_aa/180);
            
                                                           
            l_lambda[ii,jj] = (3/2)*GD_t1_depol;
            
            #GD_VI -- RH-RV/LH-LV
            
            fp22[ii,jj] = (min_sc_oc/max_sc_oc);
            
            # print(ii,jj)
            # print(num,den1,den2,den)
            # print(C11c,C12c,'SC',SC,'OC',OC,'GD',GD_t1_depol, fp22[ii,jj], 'll',l_lambda[ii][jj])

    vi_c = np.real((1 - l_lambda)*np.power(fp22, 2*l_lambda))

    return vi_c.astype(np.float32)




# Example input: List of 2D NumPy arrays
arr1 = np.random.rand(200,200).astype(np.float32)
arr2 = np.random.rand(200, 200).astype(np.float32)
arr3 = np.random.rand(200, 200).astype(np.float32)
arr4 = np.random.rand(200, 200).astype(np.float32)
# arr2 = np.random.rand(200, 200).astype(np.float32)


num_nans = 500  # Number of NaNs to insert
nan_indices = np.random.choice(arr1.size, num_nans, replace=False) 

nan_indices_2d = np.unravel_index(nan_indices, arr1.shape)

arr1[nan_indices_2d] = np.nan
arr2[nan_indices_2d] = np.nan
arr3[nan_indices_2d] = np.nan
arr4[nan_indices_2d] = np.nan


arr_list = [arr1, arr2, arr3, arr4]

# Process the arrays
# processed_arr_list = process_arrays.process_arrays(arr_list)

# # Print the results
# for i, processed_arr in enumerate(processed_arr_list):
#     print(f"Processed array {i}:\n{processed_arr}")

print("Input array statistics:")
for i, arr in enumerate(arr_list):
    print(f"Array {i}: mean={np.nanmean(arr)}, min={np.nanmin(arr)}, max={np.nanmax(arr)}")

# processed_arr_list = testpip.refined_lee.refined_lee(arr_list,7)

import time
t0 = time.time()

cprvi_cpp_out = polsartools.cprvicpp.process_chunk_cprvicpp(arr_list,5,['sdfgfd0','sdfgfd1','sdfgfd2','sdfgfd3'],-45,0)
print(f"Output array statistics: mean={np.nanmean(cprvi_cpp_out)}, min={np.nanmin(cprvi_cpp_out)}, max={np.nanmax(cprvi_cpp_out)}")
print('Time for cprvi_cpp:', time.time() - t0)
t1 = time.time()
cprvi_py_out = process_chunk_cprvi(arr_list,5,['sdfgfd0','sdfgfd1','sdfgfd2','sdfgfd3'],-45,0)
print(f"Output array statistics: mean={np.nanmean(cprvi_py_out)}, min={np.nanmin(cprvi_py_out)}, max={np.nanmax(cprvi_py_out)}")
print('Time for cprvi_py:', time.time() - t1)

print('Done!!')

are_equal = np.allclose(cprvi_cpp_out, cprvi_py_out,rtol=1e-5, atol=1e-8)
print(are_equal) 
# print(cprvi_cpp_out)
# print(cprvi_py_out)

# print(cprvi_cpp_out-cprvi_py_out)
# Print the results
# for i, processed_arr in enumerate(processed_arr_list):
#     print(f"Processed array {i}:\n{processed_arr}")
# print(processed_arr_list.shape)
# for i, processed_arr in enumerate(processed_arr_list):
    # print(f"Processed array {i}: mean={np.nanmean(processed_arr)}, min={np.nanmin(processed_arr)}, max={np.nanmax(processed_arr)}")


import os
import numpy as np
from polsartools.utils.utils import process_chunks_parallel, time_it, conv2d
from polsartools.utils.convert_matrices import C3_T3_mat

@time_it
def prvifp(infolder, outname=None, chi_in=0, psi_in=0, window_size=1,write_flag=True,max_workers=None):

    if os.path.isfile(os.path.join(infolder,"T11.bin")):
        input_filepaths = [
        os.path.join(infolder,"T11.bin"),
        os.path.join(infolder,'T12_real.bin'), os.path.join(infolder,'T12_imag.bin'),  
        os.path.join(infolder,'T13_real.bin'), os.path.join(infolder,'T13_imag.bin'),
        os.path.join(infolder,"T22.bin"),
        os.path.join(infolder,'T23_real.bin'), os.path.join(infolder,'T23_imag.bin'),  
        os.path.join(infolder,"T33.bin"),
        ]
    elif os.path.isfile(os.path.join(infolder,"C11.bin")):
        input_filepaths = [
        os.path.join(infolder,"C11.bin"),
        os.path.join(infolder,'C12_real.bin'), os.path.join(infolder,'C12_imag.bin'),  
        os.path.join(infolder,'C13_real.bin'), os.path.join(infolder,'C13_imag.bin'),
        os.path.join(infolder,"C22.bin"),
        os.path.join(infolder,'C23_real.bin'), os.path.join(infolder,'C23_imag.bin'),  
        os.path.join(infolder,"C33.bin"),
        ]

    else:
        print(f"Invalid C3 or T3 folder!!")

    output_filepaths = []
    if outname is None:
        output_filepaths.append(os.path.join(infolder, "dop_fp.tif"))
    
    process_chunks_parallel(input_filepaths, list(output_filepaths), window_size=window_size, write_flag=write_flag,
            processing_func=process_chunk_prvifp,
            block_size=(512, 512), max_workers=max_workers, 
            num_outputs=1)

def process_chunk_prvifp(chunks, window_size, input_filepaths,*args):

    if 'T11' in input_filepaths[0] and 'T22' in input_filepaths[5] and 'T33' in input_filepaths[8]:
        t11_T1 = np.array(chunks[0])
        t12_T1 = np.array(chunks[1])+1j*np.array(chunks[2])
        t13_T1 = np.array(chunks[3])+1j*np.array(chunks[4])
        t21_T1 = np.conj(t12_T1)
        t22_T1 = np.array(chunks[5])
        t23_T1 = np.array(chunks[6])+1j*np.array(chunks[7])
        t31_T1 = np.conj(t13_T1)
        t32_T1 = np.conj(t23_T1)
        t33_T1 = np.array(chunks[8])

        T_T1 = np.array([[t11_T1, t12_T1, t13_T1], 
                     [t21_T1, t22_T1, t23_T1], 
                     [t31_T1, t32_T1, t33_T1]])


    if 'C11' in input_filepaths[0] and 'C22' in input_filepaths[5] and 'C33' in input_filepaths[8]:
        C11 = np.array(chunks[0])
        C12 = np.array(chunks[1])+1j*np.array(chunks[2])
        C13 = np.array(chunks[3])+1j*np.array(chunks[4])
        C21 = np.conj(C12)
        C22 = np.array(chunks[5])
        C23 = np.array(chunks[6])+1j*np.array(chunks[7])
        C31 = np.conj(C13)
        C32 = np.conj(C23)
        C33 = np.array(chunks[8])
        C3 = np.array([[C11, C12, C13], 
                         [C21, C22, C23], 
                         [C31, C32, C33]])

        T_T1 = C3_T3_mat(C3)


    if window_size>1:
        kernel = np.ones((window_size,window_size),np.float32)/(window_size*window_size)

        t11f = conv2d(T_T1[0,0,:,:],kernel)
        t12f = conv2d(np.real(T_T1[0,1,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,1,:,:]),kernel)
        t13f = conv2d(np.real(T_T1[0,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[0,2,:,:]),kernel)
        
        t21f = np.conj(t12f) 
        t22f = conv2d(T_T1[1,1,:,:],kernel)
        t23f = conv2d(np.real(T_T1[1,2,:,:]),kernel)+1j*conv2d(np.imag(T_T1[1,2,:,:]),kernel)

        t31f = np.conj(t13f) 
        t32f = np.conj(t23f) 
        t33f = conv2d(T_T1[2,2,:,:],kernel)

        T_T1 = np.array([[t11f, t12f, t13f], [t21f, t22f, t23f], [t31f, t32f, t33f]])


    reshaped_arr = T_T1.reshape(3, 3, -1).transpose(2, 0, 1)
    det_T3 = np.linalg.det(reshaped_arr)
    # del reshaped_arr
    det_T3 = det_T3.reshape(T_T1.shape[2], T_T1.shape[3])

    trace_T3 = T_T1[0,0,:,:] + T_T1[1,1,:,:] + T_T1[2,2,:,:]
    dop_fp = np.real(np.sqrt(1-(27*(det_T3/(trace_T3**3)))))
    
    prvi = np.real((1-dop_fp)* T_T1[2,2,:,:]*0.5)  # (1-dop)*vh

    return prvi